"""SHV RPC type date and time."""

from __future__ import annotations

import datetime
import typing

from .. import SHVType
from .base import RpcType


class RpcTypeDateTime(RpcType):
    """The DateTime type representation."""

    __obj = None

    def __new__(cls) -> RpcTypeDateTime:  # noqa: D102
        if cls.__obj is None:
            cls.__obj = object.__new__(cls)
        return cls.__obj

    def __str__(self) -> str:
        return "t"

    @staticmethod
    def validate(value: SHVType) -> typing.TypeGuard[datetime.datetime]:  # noqa: D102
        return isinstance(value, datetime.datetime)


rpctype_datetime = RpcTypeDateTime()
"""The singleton for :class:`RpcTypeDateTime`."""



"""RPC File access and implementation utility."""

from __future__ import annotations

import asyncio
import binascii
import collections
import dataclasses
import datetime
import enum
import hashlib
import io
import pathlib
import types
import typing

from ..rpcapi import SHVBase
from ..rpcparam import shvarg, shvargt, shvget, shvgett, shvt
from ..value import SHVType, is_shvimap
from .access import RpcAccess
from .dir import RpcDir
from .errors import RpcInvalidParamError


@dataclasses.dataclass
class RpcFileStat:
    """The stat information for the RPC File."""

    class Type(enum.IntEnum):
        """File type defined for SHV File nodes."""

        REGULAR = 0
        MTD = 1

    tp: Type
    """Type of the SHV file node."""
    size: int
    """Size of the file in bytes."""
    page_size: int
    """Page size (ideal size and thus alignment for efficient access)."""
    access_time: datetime.datetime | None = None
    """Optional information about the latest data access."""
    mod_time: datetime.datetime | None = None
    """Optional information about the latest data modification."""
    max_write: int | None = None
    """Optional maximal size in bytes of a single write that is accepted."""
    max_read: int | None = None
    """Maximal size of the read supported by file node."""
    erase_size: int | None = None
    """Size of the page on MTD file node type."""

    class Key(enum.IntEnum):
        """Key in the stat IMap."""

        TYPE = 0
        SIZE = 1
        PAGE_SIZE = 2
        ACCESS_TIME = 3
        MOD_TIME = 4
        MAX_WRITE = 5
        MAX_READ = 6
        ERASE_SIZE = 7

    def to_shv(self) -> SHVType:
        """Convert to SHV RPC representation."""
        res: dict[int, SHVType] = {
            self.Key.TYPE: 0,
            self.Key.SIZE: self.size,
            self.Key.PAGE_SIZE: self.page_size,
        }
        if self.access_time is not None:
            res[self.Key.ACCESS_TIME] = self.access_time
        if self.mod_time is not None:
            res[self.Key.MOD_TIME] = self.mod_time
        if self.max_write is not None:
            res[self.Key.MAX_WRITE] = self.max_write
        return res

    @classmethod
    def from_shv(cls, value: SHVType) -> RpcFileStat:
        """Create from SHV RPC representation."""
        if not is_shvimap(value):
            raise ValueError("Expected Map.")
        access_time = shvget(value, cls.Key.ACCESS_TIME, None)
        access_time = (
            None if access_time is None else shvt(access_time, datetime.datetime)
        )
        mod_time = shvget(value, cls.Key.MOD_TIME, None)
        mod_time = None if mod_time is None else shvt(mod_time, datetime.datetime)
        max_write = shvget(value, cls.Key.MAX_WRITE, None)
        max_write = None if max_write is None else shvt(max_write, int)
        max_read = shvget(value, cls.Key.MAX_READ, None)
        max_read = None if max_read is None else shvt(max_read, int)
        erase_size = shvget(value, cls.Key.ERASE_SIZE, None)
        erase_size = None if erase_size is None else shvt(erase_size, int)
        return cls(
            tp=cls.Type(shvgett(value, cls.Key.TYPE, int, 0)),
            size=shvgett(value, cls.Key.SIZE, int, 0),
            page_size=shvgett(value, cls.Key.PAGE_SIZE, int, 128),
            access_time=access_time,
            mod_time=mod_time,
            max_write=max_write,
            max_read=max_read,
            erase_size=erase_size,
        )

    @classmethod
    def for_path(cls, path: pathlib.Path | str) -> RpcFileStat:
        """Create stat information for the existing file.

        This doesn't support MTD type!

        :param path: Path to the file for which stat should be provided.
        :return: The RPC stat object.
        :raises FileNotFoundError: If there is no such file or if path exists
          but doesn't lead to the file.
        """
        if isinstance(path, str):
            path = pathlib.Path(path)
        if not path.is_file():
            raise FileNotFoundError("Not a regular file")
        stat = path.stat()
        return cls(
            cls.Type.REGULAR,
            stat.st_size,
            stat.st_blksize,
            datetime.datetime.fromtimestamp(stat.st_atime),
            datetime.datetime.fromtimestamp(stat.st_mtime),
        )


def _ifloor(val: int, mult: int) -> int:
    return val - (val % mult)


def _iceil(val: int, mult: int) -> int:
    return val + (mult - (val % mult))


class RpcFile:
    """RPC file accessed over :class:`shv.rpcapi.SHVBase`.

    :param client: The :class:`shv.rpcapi.SHVBase` based client used to
      communicate.
    :param path: SHV path to the file node.
    :param buffered: If reading and writing should be backed by local buffer to
      transfer data in chunks suggested by the device.
    :param append: If writing should be done through append instead of write.
    """

    class Flag(enum.Flag):
        """File mode and operation selection flags."""

        APPEND = enum.auto()
        """Write in append mode.

        The append mode changes the used write operation from 'write' to
        'append'. That way writes are always at the end of the file regardless
        of the current offset.
        """
        R_BUFFERED = enum.auto()
        """Buffered read.

        Read operations are buffered. Reads are performed in page size blocks.
        """
        W_BUFFERED = enum.auto()
        """Buffered write.

        Write operations are buffered. The exact behavior depends if ``APPEND`
        is used as well or not. Writes are buffered up to the maximum write size
        signaled by file (or page size of not provided) if ``APPEND`` is not
        used. With ``APPEND`` all writes are buffered until
        :meth:`RpcFile.flush` is awaited
        """
        BUFFERED = R_BUFFERED | W_BUFFERED
        """The combination of the read and write buffering."""

    def __init__(
        self,
        client: SHVBase,
        path: str,
        flags: RpcFile.Flag = Flag.BUFFERED,
    ) -> None:
        self.client: SHVBase = client
        """The client used to access the RPC file."""
        self._path = path
        self._flags = flags
        self._offset: int = 0
        self._read_offset: int = 0
        self._read_buffer: bytes = b""
        self._write_offset: int = 0
        self._write_buffer: bytearray = bytearray()
        self.__page_size: int | None = None
        self.__max_write: int | None = None

    @property
    def path(self) -> str:
        """SHV path to the file this object is used to access."""
        return self._path

    @property
    def offset(self) -> int:
        """Offset in bytes from the start of the file.

        This is used as offset for subsequent reads and write operations (with
        exception if append mode is used).

        The offset gets modified automatically by reading and writing. You can
        also set it to perform seek.
        """
        return self._offset

    @offset.setter
    def offset(self, value: int) -> None:
        if value < 0:
            raise ValueError("Offset can't be negative")
        self._offset = value

    async def __aenter__(self) -> RpcFile:
        self._offset = 0
        return self

    async def __aexit__(
        self,
        exc_type: type[Exception],
        exc_value: Exception,
        traceback: types.TracebackType,
    ) -> None:
        await self.flush()

    async def __aiter__(self) -> RpcFile:
        return self

    async def __anext__(self) -> bytes:
        res = await self.readuntil()
        if not res:
            raise StopAsyncIteration
        return res

    async def readable(self) -> bool:
        """Check if file is readable."""
        return await self.client.dir_exists(self._path, "read")

    async def writable(self) -> bool:
        """Check if file is writable with selected write method."""
        return await self.client.dir_exists(
            self._path, "append" if self.Flag.APPEND in self._flags else "write"
        )

    async def resizable(self) -> bool:
        """Check if file is resizable."""
        return await self.client.dir_exists(self._path, "truncate")

    async def stat(self) -> RpcFileStat:
        """Get the file stat info."""
        return RpcFileStat.from_shv(await self.client.call(self._path, "stat"))

    async def size(self) -> int:
        """Get the file size."""
        res = await self.client.call(self._path, "size")
        return res if isinstance(res, int) else 0

    async def crc32(self, offset: int | None = 0, size: int | None = None) -> int:
        """Calculate CRC32 checksum of the data.

        :param offset: offset from the file start or current file offset if
          ``None`` is passed.
        :param size: number of bytes since offset used to calculate checksum or
          all of them up to the end of the file in case ``None`` is passed.
        :return: Calculated CRC32 value.
        """
        if offset == 0 and size is None:
            param = None
        else:
            param = [self._offset if offset is None else offset, size]
        res = await self.client.call(self._path, "crc", param)
        return res if isinstance(res, int) else 0

    async def sha1(self, offset: int | None = 0, size: int | None = None) -> bytes:
        """Calculate SHA1 checksum of the data.

        :param offset: offset from the file start or current file offset if
          ``None`` is passed.
        :param size: number of bytes since offset used to calculate checksum or
          all of them up to the end of the file in case ``None`` is passed.
        :return: Calculated SHA1 value.
        """
        if offset == 0 and size is None:
            param = None
        else:
            param = [self._offset if offset is None else offset, size]
        res = await self.client.call(self._path, "sha1", param)
        return res if isinstance(res, bytes) else bytes(20)

    async def __fetch_stat_info(self) -> None:
        stat = await self.stat()
        self.__page_size = stat.page_size
        self.__max_write = stat.max_write or stat.page_size

    async def _page_size(self) -> int:
        """Get the ideal page size."""
        if self.__page_size is None:
            await self.__fetch_stat_info()
        assert self.__page_size is not None
        return self.__page_size

    async def _max_write(self) -> int:
        """Get the maximum write size."""
        if self.__max_write is None:
            await self.__fetch_stat_info()
        assert self.__max_write is not None
        return self.__max_write

    async def _unbuf_read(self, offset: int, size: int = -1) -> bytearray:
        size = size or -1  # Just so we can use 0 in here
        page_size = await self._page_size()
        result = bytearray()
        while size != 0:
            cnt = size if 0 < size < page_size else page_size
            data = await self.client.call(self._path, "read", [offset, cnt])
            if not isinstance(data, bytes):
                raise Exception("Invalid result type from file read")
            if not data:
                break
            result += data
            offset += len(data)
            if size > 0:
                size -= len(data)
        return result

    async def read(self, size: int = -1) -> bytes:
        """Read at most given number of bytes.

        :param size: Maximum number of bytes to read. The value 0 or less is the
          no size limit.
        """
        if self.Flag.R_BUFFERED not in self._flags:
            result = await self._unbuf_read(self._offset, size)
            self._offset += len(result)
            return bytes(result)

        # Flush the write buffer if read overlaps with it
        if (self.Flag.W_BUFFERED | self.Flag.APPEND) in self._flags:
            if (self._write_offset <= (self._offset + (size if size > 0 else 0))) or (
                (self._write_offset + len(self._write_buffer)) >= self._offset
            ):
                await self.flush()

        page_size = await self._page_size()
        result = bytearray()
        if (
            self.Flag.R_BUFFERED in self._flags
            and self._read_offset <= self._offset <= self._read_offset + page_size
        ):
            bstart = self._offset - self._read_offset
            bend = bstart + size if size > 0 else None
            result += self._read_buffer[self._offset - self._read_offset : bend]
            if size > 0 and len(result) == size:  # Served all from buffer
                self._offset += len(result)
                return bytes(result)
            # Note: We must try to fetch if we don't have full page to ensure
            # that we can read files that are being appended to.
        pstart = _ifloor(self._offset + len(result), page_size)
        data = await self._unbuf_read(
            pstart, _iceil(size - len(result), page_size) if size > 0 else -1
        )
        poff = _ifloor(len(data), page_size)
        self._read_buffer = data[poff:]
        self._read_offset = pstart + poff
        doff = self._offset + len(result) - pstart
        result += data[doff : (doff + size - len(result)) if size > 0 else None]
        self._offset += len(result)
        return result

    async def readuntil(self, separator: bytes = b"\n") -> bytes:
        """Read and return one line from the file from the current offset."""
        # TODO with buffering we should fetch page and look for separator in it
        result = bytearray()
        # TODO We could be smarter here and read len(separator) if result is
        # not ending with possible partial sequence
        while separator not in result:
            data = await self._unbuf_read(self._offset, 1)
            if not data:
                break
            result += data
            self._offset += len(data)
        return bytes(result)

    async def _page_write(self, offset: int, data: bytes | bytearray) -> None:
        max_write = await self._max_write()
        while data:
            towrite = bytes(data[:max_write])
            data = data[max_write:]
            await self.client.call(self._path, "write", [offset, towrite])
            offset += len(towrite)

    async def write(self, data: bytes | bytearray) -> None:
        """Write data to the file on the current offset."""
        if self.Flag.W_BUFFERED not in self._flags:
            if self.Flag.APPEND in self._flags:
                await self.client.call(self._path, "append", data)
            else:
                await self._page_write(self._offset, data)
                self._offset += len(data)
            return

        if self.Flag.APPEND in self._flags:
            self._write_buffer += data
            return

        if (
            self._write_buffer
            and self._write_offset
            <= self._offset
            <= self._write_offset + len(self._write_buffer)
        ):  # Adding or modifying the existing buffer
            self._write_buffer[self._offset - self._write_offset :] = data
        else:
            if self._write_buffer:
                # We are attempting to write to the different location
                await self._page_write(self._write_offset, self._write_buffer)
            self._write_buffer[:] = data
            self._write_offset = self._offset

        page_size = await self._max_write()
        if len(self._write_buffer) > page_size:
            # Write full pages immediately
            size = _ifloor(len(self._write_buffer), page_size)
            await self._page_write(self._write_offset, self._write_buffer[:size])
            del self._write_buffer[:size]

        self._offset += len(data)

    async def flush(self) -> None:
        """Flush data buffered for write by previous :meth:`write` calls.

        This is required only if buffering is enabled. It otherwise does
        nothing.

        Note that in combination of buffering and append mode this must be
        called because write will never flush buffer on its own.
        """
        if self.Flag.W_BUFFERED not in self._flags:
            return
        if self.Flag.APPEND in self._flags:
            await self.client.call(self._path, "append", self._write_buffer)
        else:
            await self._page_write(self._write_offset, self._write_buffer)
        self._write_buffer = bytearray()

    async def truncate(self, size: int | None = None) -> None:
        """Truncate file size.

        :param size: Truncate to this specific size or to the current offset in
          case of ``None``.
        """
        await self.client.call(
            self._path, "truncate", size if size is not None else self._offset
        )


@dataclasses.dataclass
class FileProvider:
    """Implementation helper that provides local files over SHV RPC.

    This should be used from :class:`shv.rpcapi.SHVBase` based classes. The
    appropriate file provide is instanciated and for the appropriate file nodes
    all `dir` and other method calls are delefated to this helper.
    """

    access_read: RpcAccess = RpcAccess.READ
    """Access level for ``read`` method."""
    access_write: RpcAccess | None = RpcAccess.WRITE
    """Access level for ``write`` method."""
    access_truncate: RpcAccess | None = RpcAccess.WRITE
    """Access level for ``truncate`` method."""
    access_append: RpcAccess | None = RpcAccess.WRITE
    """Access level for ``append`` method."""

    def dir(self) -> collections.abc.Iterator[RpcDir]:
        """Provide method descriptions for this file access.

        This should be called from :meth:`shv.rpcapi.SHVBase._dir`
        implementation for file nodes.

        :return: Iterator over method descriptions appropriate for this
          provider.
        """
        yield RpcDir.getter("stat", "n", "!stat", self.access_read)
        yield RpcDir.getter("size", "n", "i(0,)", self.access_read)
        yield RpcDir(
            "crc",
            param="[i(0,):offset,i(0,)|n:size]|n",
            result="u(>32)",
            access=self.access_read,
        )
        yield RpcDir(
            "sha1",
            param="[i(0,):offset,i(0,)|n:size]|n",
            result="b(20)",
            access=self.access_read,
        )
        yield RpcDir(
            "read",
            RpcDir.Flag.LARGE_RESULT_HINT,
            "[i(0,):offset,i(0,)size]",
            "b",
            self.access_read,
        )
        if self.access_write is not None:
            yield RpcDir(
                "write", param="[i(0,):offset,b:data]", access=self.access_write
            )
        if self.access_truncate is not None:
            yield RpcDir("truncate", param="i(0,)", access=self.access_truncate)
        if self.access_append is not None:
            yield RpcDir("append", param="b", access=self.access_append)

    async def method_call(
        self, path: pathlib.Path, request: SHVBase.Request
    ) -> SHVType:
        """File access methods implementation.

        This should be called from :meth:`shv.rpcapi.SHVBase._method_call`
        implementation for file nodes.

        :param path: Path to the file that should be accessed. This is not SHV
          path but rather a real local file path.
        :param request: The request parameter from
          :meth:`shv.rpcapi.SHVBase._method_call`.
        :return: The value that should be provided as result of the method call.
        :raises FileNotFoundError: If file pointed to by ``path`` doesn't
          exists.
        :raises RpcInvalidParamError: If parameter is invalid.
        :raises NotImplementedError: If requested method is not implemented or
          too low access level was provided.
        """
        match request.method:
            case "stat" if request.access >= self.access_read:
                return RpcFileStat.for_path(path).to_shv()
            case "size" if request.access >= self.access_read:
                stat = await asyncio.get_running_loop().run_in_executor(None, path.stat)
                return stat.st_size
            case "crc" if request.access >= self.access_read:
                offset = shvargt(request.param, 0, int, 0)
                size = shvarg(request.param, 1, None)
                if size is not None:
                    size = shvt(size, int)
                crc = binascii.crc32(b"")
                with path.open("rb") as file:
                    file.seek(offset)
                    while size is None or size > 0:
                        data = file.read(
                            io.DEFAULT_BUFFER_SIZE
                            if size is None
                            else min(size, io.DEFAULT_BUFFER_SIZE)
                        )
                        if not data:
                            break
                        crc = binascii.crc32(data, crc)
                        if size is not None:
                            size -= len(data)
                return crc
            case "sha1" if request.access >= self.access_read:
                offset = shvargt(request.param, 0, int, 0)
                size = shvarg(request.param, 1, None)
                if size is not None:
                    size = shvt(size, int)
                hash = hashlib.sha1()  # noqa PLR6301
                with path.open("rb") as file:
                    file.seek(offset)
                    while size is None or size > 0:
                        data = file.read(
                            io.DEFAULT_BUFFER_SIZE
                            if size is None
                            else min(size, io.DEFAULT_BUFFER_SIZE)
                        )
                        if not data:
                            break
                        hash.update(data)
                        if size is not None:
                            size -= len(data)
                return hash.digest()
            case "read" if request.access >= self.access_read:
                offset = shvargt(request.param, 0, int)
                size = shvargt(request.param, 1, int)
                with path.open("rb") as file:
                    file.seek(offset)
                    return file.read(size)
            case "write" if (
                self.access_write is not None and request.access >= self.access_write
            ):
                offset = shvargt(request.param, 0, int)
                data = shvargt(request.param, 1, bytes)
                stat = await asyncio.get_running_loop().run_in_executor(None, path.stat)
                if self.access_truncate is None and (offset + len(data)) > stat.st_size:
                    raise RpcInvalidParamError(
                        "Write beyond the file boundary is not possible"
                    )
                with path.open("r+b") as file:
                    file.seek(offset)
                    file.write(data)
                return None
            case "truncate" if (
                self.access_truncate is not None
                and request.access >= self.access_truncate
            ):
                size = shvt(request.param, int)
                with path.open("wb") as file:
                    file.truncate(size)
                return None
            case "append" if (
                self.access_append is not None and request.access >= self.access_append
            ):
                data = shvt(request.param, bytes)
                with path.open("ab") as file:
                    file.write(data)
                return None
        raise NotImplementedError


FileProviderRO: typing.Final = FileProvider(
    access_read=RpcAccess.READ,
    access_write=None,
    access_truncate=None,
    access_append=None,
)
"""RPC File provider for read-only file access."""
FileProviderRW: typing.Final = FileProvider(
    access_read=RpcAccess.READ,
    access_write=RpcAccess.WRITE,
    access_truncate=RpcAccess.WRITE,
    access_append=RpcAccess.WRITE,
)
"""RPC File provider for read-write access to the files."""
FileProviderFixedSize: typing.Final = FileProvider(
    access_read=RpcAccess.READ,
    access_write=RpcAccess.WRITE,
    access_truncate=None,
    access_append=None,
)
"""RPC File provider for read-write access to the files of fixed size."""
FileProviderAppend: typing.Final = FileProvider(
    access_read=RpcAccess.READ,
    access_write=None,
    access_truncate=None,
    access_append=RpcAccess.WRITE,
)
"""RPC File provider that aprovides write access but only through appends."""



"""Parse the SHV RPC type string."""

from __future__ import annotations

import decimal
import re

from .any import RpcTypeAny
from .base import RpcType
from .bitfield import RpcTypeBitfield
from .blob import RpcTypeBlob
from .bool import rpctype_bool
from .datetime import RpcTypeDateTime
from .decimal import RpcTypeDecimal
from .double import RpcTypeDouble
from .enum import RpcTypeEnum
from .imap import RpcTypeIMap
from .integer import RpcTypeInteger
from .keystruct import RpcTypeKeyStruct
from .list import RpcTypeList
from .map import RpcTypeMap
from .null import rpctype_null
from .oneof import RpcTypeOneOf
from .standard import (
    rpctype_alert,
    rpctype_clientinfo,
    rpctype_dir,
    rpctype_exchange_p,
    rpctype_exchange_r,
    rpctype_exchange_v,
    rpctype_getlog_p,
    rpctype_getlog_r,
    rpctype_history_records,
    rpctype_stat,
)
from .string import RpcTypeString
from .struct import RpcTypeStruct
from .tuple import RpcTypeTuple
from .unsigned import RpcTypeUnsigned

_decimal_re = re.compile(r"-?(?:0|(?:[1-9][0-9]*))?(?:\.[0-9]+)?")


def rpctype_parse(rpctype: str) -> RpcType:
    """Parse the given value as RPC Type string.

    :param rpctype: The string representation of the RPC type hint.
    :return: The object representation.
    :raises RpcTypeParseError: In case type hint has invalid format.
    """
    i = 0

    def hasc(expected: str, whitelist: bool = True) -> str | None:
        nonlocal i
        if len(rpctype) > i and (
            expected is None or ((rpctype[i] in expected) is whitelist)
        ):
            i += 1
            return rpctype[i - 1]
        return None

    def check(expected: str) -> None:
        nonlocal i
        if not rpctype.startswith(expected, i):
            raise RpcTypeParseError(f"Expected '{expected}' at {i}")
        i += len(expected)

    def _parse_int() -> int:
        value = 0
        while c := hasc("123456789" + ("0" if value else "")):
            value = value * 10 + ord(c) - ord("0")
        return value

    def parse_number(positive: bool = False) -> int | None:
        if hasc("0"):
            return 0
        positive = positive or not hasc("-")
        mod = hasc("^>")
        value = _parse_int()
        return (
            (2**value if mod == "^" else 2**value - 1 if mod == ">" else value)
            * (1 if positive else -1)
            if value
            else None
        )

    def parse_number_required(positive: bool = False) -> int:
        previ = i
        if (res := parse_number(positive)) is not None:
            return res
        raise RpcTypeParseError(f"Expected number at {previ}")

    def parse_decimal_number() -> decimal.Decimal | None:
        nonlocal i
        match = _decimal_re.match(rpctype[i:])
        if not match or not match.group(0):
            return None
        endi = i + len(match.group(0))
        result = decimal.Decimal(rpctype[i:endi])
        i = endi
        return result

    def parse_text() -> str:
        return "".join(iter(lambda: hasc("[]{}():,|", False), None))

    def parse_text_required() -> str:
        if result := parse_text():
            return result
        raise RpcTypeParseError(f"Text expected at {i}")

    def parse_integer() -> RpcTypeInteger:
        imin, imax = None, None
        if hasc("("):
            imin = parse_number()
            check(",")
            imax = parse_number()
            check(")")
        return RpcTypeInteger(imin, imax, parse_text())

    def parse_unsigned() -> RpcTypeUnsigned:
        imin, imax = None, None
        if hasc("("):
            imax = parse_number(True)
            if hasc(","):
                imin = imax
                imax = parse_number(True)
            check(")")
        return RpcTypeUnsigned(imin or 0, imax, parse_text())

    def parse_decimal() -> RpcTypeDecimal:
        dmin, dmax, dpres = None, None, None
        if hasc("("):
            dmin = parse_decimal_number()
            check(",")
            dmax = parse_decimal_number()
            if hasc(","):
                dpres = parse_number()
            check(")")
        return RpcTypeDecimal(dmin, dmax, dpres, parse_text())

    def parse_enum() -> RpcTypeEnum:
        items = {}
        index = 0
        while True:
            key = parse_text_required()
            if hasc(":"):
                index = parse_number_required()
            if index in items:
                raise RpcTypeParseError(f"Reuse of the integer key {index}")
            items[index] = key
            index += 1
            if hasc("]"):
                break
            check(",")
        return RpcTypeEnum(items)

    def parse_size() -> tuple[int, int | None]:
        imin, imax = None, None
        if hasc("("):
            imin = parse_number(True)
            imax = parse_number(True) if hasc(",") else imin
            check(")")
        return imin or 0, imax

    def parse_list_tuple() -> RpcTypeList | RpcTypeTuple:
        tp = parse()
        if hasc("]"):  # List
            return RpcTypeList(tp, *parse_size())
        # Tuple
        ivalues = []
        while True:
            check(":")
            ivalues.append((tp, parse_text_required()))
            if not hasc(","):
                break
            tp = parse()
        check("]")
        return RpcTypeTuple(*ivalues)

    def parse_imap_struct() -> RpcTypeIMap | RpcTypeStruct:
        tp = parse()
        if hasc("}"):  # IMap
            return RpcTypeIMap(tp)
        # Struct
        values = {}
        index = 0
        while True:
            check(":")
            key = parse_text_required()
            if hasc(":"):
                index = parse_number_required()
            if index in values:
                raise RpcTypeParseError(f"Reuse of the integer key {index}")
            values[index] = (tp, key)
            index += 1
            if hasc("}"):
                break
            check(",")
            tp = parse()
        return RpcTypeStruct(values)

    def parse_map_keystruct() -> RpcTypeMap | RpcTypeKeyStruct:
        tp = parse()
        if hasc("}"):  # Map
            return RpcTypeMap(tp)
        # KeyStruct
        mvalues = {}
        while True:
            check(":")
            key = parse_text_required()
            if key in mvalues:
                raise RpcTypeParseError(f"Reuse of the key '{key}'")
            mvalues[key] = tp
            if hasc("}"):
                break
            check(",")
            tp = parse()
        return RpcTypeKeyStruct(mvalues)

    def parse_bitfield() -> RpcTypeBitfield:
        bitems = []
        ikey = 0
        while True:
            tp = parse()
            tpsiz = RpcTypeBitfield.bitsize(tp)
            check(":")
            key = parse_text_required()
            if hasc(":"):
                ikey = parse_number_required()
            bitems.append((ikey, tp, key))
            ikey += tpsiz or 0
            if hasc("]"):
                break
            check(",")
        return RpcTypeBitfield(*bitems)

    def parse() -> RpcType:
        res: RpcType | None = None
        match hasc("nbiufdsxt[{!?"):
            case "n":  # Null
                res = rpctype_null
            case "b":  # Bool
                res = rpctype_bool
            case "i":
                match hasc("[{"):
                    case "[":  # Enum
                        res = parse_enum()
                    case "{":  # IMap or Struct
                        res = parse_imap_struct()
                    case _:  # Integer
                        res = parse_integer()
            case "u":
                match hasc("["):
                    case "[":  # Bitfield
                        res = parse_bitfield()
                    case _:  # Unsigned integer
                        res = parse_unsigned()
            case "f":  # Double (floating point number)
                res = RpcTypeDouble(parse_text())
            case "d":  # Decimal
                res = parse_decimal()
            case "s":  # String
                res = RpcTypeString(*parse_size())
            case "x":  # Blob
                res = RpcTypeBlob(*parse_size())
            case "t":  # DateTime
                res = RpcTypeDateTime()
            case "[":  # List or Tuple
                res = parse_list_tuple()
            case "{":
                res = parse_map_keystruct()
            case "!":  # Standard types
                match parse_text_required():
                    case "dir":
                        res = rpctype_dir
                    case "alert":
                        res = rpctype_alert
                    case "clientInfo":
                        res = rpctype_clientinfo
                    case "stat":
                        res = rpctype_stat
                    case "exchangeP":
                        res = rpctype_exchange_p
                    case "exchangeR":
                        res = rpctype_exchange_r
                    case "exchangeV":
                        res = rpctype_exchange_v
                    case "getLogP":
                        res = rpctype_getlog_p
                    case "getLogR":
                        res = rpctype_getlog_r
                    case "historyRecords":
                        res = rpctype_history_records
                    case tpname:
                        raise RpcTypeParseError(f"Invalid standard type '{tpname}'")
            case "?":
                name = ""
                if hasc("("):
                    name = parse_text()
                    check(")")
                res = RpcTypeAny(name)
        if res is None:
            raise RpcTypeParseError(
                f"Invalid character '{rpctype[i]}' at {i}"
                if len(rpctype) > i
                else f"Missing character at {i}"
            )
        if hasc("|"):
            res = RpcTypeOneOf(res, parse())
        return res

    try:
        result = parse()
    except ValueError as exc:
        raise RpcTypeParseError(*exc.args) from exc
    if i != len(rpctype):
        raise RpcTypeParseError(f"Unexpected character '{rpctype[i]}' at {i}")
    return result


class RpcTypeParseError(ValueError):
    """The RPC Type parse error."""



"""Check various clients and servers implementing a different link layers."""

from __future__ import annotations

import asyncio
import collections.abc
import contextlib
import dataclasses
import io
import logging
import os
import pathlib
import pty
import random
import select
import ssl
import threading

import pytest

from shv.rpcmessage import RpcMessage
from shv.rpctransport import (
    SHVCAN,
    RpcClient,
    RpcClientCAN,
    RpcClientPipe,
    RpcClientTCP,
    RpcClientTTY,
    RpcClientUnix,
    RpcClientWebSockets,
    RpcServerCAN,
    RpcServerTCP,
    RpcServerTTY,
    RpcServerUnix,
    RpcServerWebSockets,
    RpcServerWebSocketsUnix,
)

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class PTYPort:
    """PTY based exchange port."""

    port1: str
    port2: str
    event: threading.Event

    @classmethod
    @contextlib.contextmanager
    def new(cls) -> collections.abc.Iterator[PTYPort]:
        """Context function that provides PTY based ports.

        :return: Tuple with two strings containing paths to the pty ports and
          event that can be cleared to pause data exchange and set to resume it
          again.
        """
        pty1_master, pty1_slave = pty.openpty()
        pty2_master, pty2_slave = pty.openpty()
        event = threading.Event()
        event.set()
        thread = threading.Thread(
            target=cls._exchange, args=(pty1_master, pty2_master, event)
        )
        thread.start()
        yield cls(os.ttyname(pty1_slave), os.ttyname(pty2_slave), event)
        os.close(pty1_slave)
        os.close(pty2_slave)
        thread.join()

    @staticmethod
    def _exchange(fd1: int, fd2: int, event: threading.Event) -> None:
        p = select.poll()
        p.register(fd1, select.POLLIN | select.POLLPRI)
        p.register(fd2, select.POLLIN | select.POLLPRI)
        fds = 2
        while fds:
            for fd, pevent in p.poll():
                event.wait()
                if pevent & select.POLLHUP:
                    p.unregister(fd)
                    fds -= 1
                if pevent & select.POLLIN:
                    data = os.read(fd, io.DEFAULT_BUFFER_SIZE)
                    os.write(fd1 if fd is fd2 else fd2, data)
                if pevent & select.POLLERR:
                    logger.error("Error detected in ptycopy")
                    return


@pytest.fixture(name="pty")
def fixture_pty():
    with PTYPort.new() as port:
        yield port


class Link:
    """Generic tests applied to link layers."""

    async def test_msg_a(self, clients):
        msg = RpcMessage.request(".app", "dir")
        await clients[0].send(msg)
        assert await clients[1].receive() == msg

    async def test_msg_b(self, clients):
        msg = RpcMessage.signal(".app")
        await clients[1].send(msg)
        assert await clients[0].receive() == msg

    @pytest.mark.parametrize("a,b", ((0, 1), (1, 0)))
    async def test_reset(self, clients, a, b):
        await clients[a].reset()
        assert await clients[b].receive() == RpcClient.Control.RESET


class ServerLink(Link):
    """Additional tests for server-client connections."""

    @pytest.mark.parametrize("cdisc,crecv", ((0, 1), (1, 0)))
    async def test_eof_receive(self, clients, cdisc, crecv):
        clients[cdisc].disconnect()
        await clients[cdisc].wait_disconnect()
        assert not clients[cdisc].connected
        with pytest.raises(EOFError):
            await clients[crecv].receive()
        assert not clients[crecv].connected

    @pytest.mark.parametrize("cdisc,crecv", ((0, 1), (1, 0)))
    async def test_eof_send(self, clients, cdisc, crecv):
        clients[cdisc].disconnect()
        await clients[cdisc].wait_disconnect()
        with pytest.raises(EOFError):
            await clients[crecv].send(RpcMessage.request(".app", "ping"))
        assert not clients[crecv].connected

    async def test_reconnect(self, clients, server):
        clients[1].disconnect()
        await clients[1].wait_disconnect()
        await clients[1].reset()
        server_client = await server[1].get()
        msg = RpcMessage.request("foo", "ls")
        await clients[1].send(msg)
        assert await server_client.receive() == msg
        server_client.disconnect()
        await server_client.wait_disconnect()


class TestPipe(Link):
    """Check that we can work over Unix pipe pair."""

    @pytest.fixture(name="clients")
    async def fixture_clients(self):
        client1, client2 = await RpcClientPipe.open_pair()
        yield client1, client2
        client1.disconnect()
        client2.disconnect()
        await client1.wait_disconnect()
        await client2.wait_disconnect()


class TestTCP(ServerLink):
    """Check that TCP/IP transport protocol works."""

    @pytest.fixture(name="ssl")
    def fixture_ssl(self):
        return None, None

    @pytest.fixture(name="server")
    async def fixture_server(self, port, ssl):
        queue = asyncio.Queue()
        server = RpcServerTCP(queue.put, "localhost", port, ssl=ssl[1])
        await server.listen()
        yield server, queue
        server.close()
        await server.wait_closed()

    @pytest.fixture(name="clients")
    async def fixture_clients(self, server, port, ssl):
        client = await RpcClientTCP.connect("localhost", port, ssl=ssl[0])
        server_client = await server[1].get()
        yield server_client, client
        client.disconnect()
        server_client.disconnect()
        await client.wait_disconnect()
        await server_client.wait_disconnect()

    async def test_before_connect(self, port) -> None:
        client = RpcClientTCP("localhost", port)
        with pytest.raises(EOFError):
            await client.receive()
        with pytest.raises(EOFError):
            await client.send(RpcMessage.request(".app", "ping"))
        client.disconnect()
        await client.wait_disconnect()

    def test_secure(self, clients):
        assert not clients[0].secure
        assert not clients[1].secure


class TestSSL(TestTCP):
    """Check that TCP/IP SSL transport protocol works."""

    @pytest.fixture(name="ssl")
    def fixture_ssl(self):
        ca = pathlib.Path(__file__).parent / "ca"

        client = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        client.load_verify_locations(cafile=ca / "ca.crt")
        client.check_hostname = False
        client.verify_mode = ssl.CERT_REQUIRED

        server = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        server.load_cert_chain(certfile=ca / "server.crt", keyfile=ca / "server.key")
        server.load_verify_locations(cafile=ca / "ca.crt")
        server.verify_mode = ssl.CERT_OPTIONAL

        return client, server

    def test_secure(self, clients):
        assert clients[0].secure
        assert clients[1].secure


class TestUnix(ServerLink):
    """Check that Unix transport protocol works."""

    @pytest.fixture(name="sockpath")
    def fixture_sockpath(self, tmp_path):
        return tmp_path / "s"

    @pytest.fixture(name="server")
    async def fixture_server(self, sockpath):
        queue = asyncio.Queue()
        server = RpcServerUnix(queue.put, sockpath)
        await server.listen()
        yield server, queue
        server.close()
        await server.wait_closed()

    @pytest.fixture(name="clients")
    async def fixture_clients(self, server, sockpath):
        client = await RpcClientUnix.connect(sockpath)
        server_client = await server[1].get()
        yield server_client, client
        server_client.disconnect()
        client.disconnect()
        await server_client.wait_disconnect()
        await client.wait_disconnect()


class TestTTY(Link):
    """Check that we can work over TTY port transport protocol."""

    @pytest.fixture(name="clients")
    async def fixture_clients(self, pty):
        # We pause data exchange to resolve race condition between terminal
        # setup and reset message sending. This ensures that ports are
        # configured to raw mode before reset message is exchanged.
        pty.event.clear()
        client1 = await RpcClientTTY.connect(pty.port1)
        client2 = await RpcClientTTY.connect(pty.port2)
        pty.event.set()

        # Flush reset sent by clients
        assert await client1.receive() is RpcClient.Control.RESET
        assert await client2.receive() is RpcClient.Control.RESET

        yield client1, client2

        client1.disconnect()
        client2.disconnect()
        await client1.wait_disconnect()
        await client2.wait_disconnect()

    async def test_escapes(self, clients):
        msg = RpcMessage.request("prop", "set", b"1\xa2\xa3\xa4\xa5\xaa2")
        await clients[0].send(msg)
        assert await clients[1].receive() == msg


class TestTTYServer(Link):
    """Check that we can work over TTY port transport protocol and its server."""

    @pytest.fixture(name="server")
    async def fixture_server(self, pty):
        pty.event.clear()
        queue = asyncio.Queue()
        server = RpcServerTTY(queue.put, pty.port1)
        await server.listen()
        yield server, queue
        server.terminate()
        await server.wait_terminated()

    @pytest.fixture(name="clients")
    async def fixture_clients(self, server, pty):
        client = await RpcClientTTY.connect(pty.port2)
        server_client = await server[1].get()
        pty.event.set()
        assert await server_client.receive() is RpcClient.Control.RESET
        assert await client.receive() is RpcClient.Control.RESET
        yield server_client, client
        client.disconnect()
        server_client.disconnect()
        await client.wait_disconnect()
        await server_client.wait_disconnect()


class TestWebSockets(ServerLink):
    """Check that WebSockets transport protocol works."""

    @pytest.fixture(name="server")
    async def fixture_server(self, port):
        queue = asyncio.Queue()
        server = RpcServerWebSockets(queue.put, "localhost", port)
        await server.listen()
        yield server, queue
        server.close()
        await server.wait_closed()

    @pytest.fixture(name="clients")
    async def fixture_clients(self, server, port):
        client = await RpcClientWebSockets.connect("localhost", port)
        server_client = await server[1].get()
        await asyncio.sleep(0)
        yield server_client, client
        client.disconnect()
        server_client.disconnect()
        await client.wait_disconnect()
        await server_client.wait_disconnect()

    async def test_before_connect(self, port) -> None:
        client = RpcClientWebSockets("localhost", port)
        with pytest.raises(EOFError):
            await client.receive()
        with pytest.raises(EOFError):
            await client.send(RpcMessage.request(".app", "ping"))
        client.disconnect()
        await client.wait_disconnect()


class TestWebSocketsUnix(ServerLink):
    """Check that WebSockets Unix transport protocol works."""

    @pytest.fixture(name="sockpath")
    def fixture_sockpath(self, tmp_path):
        return tmp_path / "s"

    @pytest.fixture(name="server")
    async def fixture_server(self, sockpath):
        queue = asyncio.Queue()
        server = RpcServerWebSocketsUnix(queue.put, sockpath)
        await server.listen()
        yield server, queue
        server.close()
        await server.wait_closed()

    @pytest.fixture(name="clients")
    async def fixture_clients(self, server, sockpath):
        client = await RpcClientWebSockets.connect(sockpath)
        server_client = await server[1].get()
        yield server_client, client
        server_client.disconnect()
        client.disconnect()
        await server_client.wait_disconnect()
        await client.wait_disconnect()


class TestCAN(ServerLink):
    """Check that CAN Bus transport protocol works."""

    @pytest.fixture(name="server")
    async def fixture_server(self, port):
        queue = asyncio.Queue()
        server = RpcServerCAN(queue.put, SHVCAN.virtualcan("test"), 42)
        await server.listen()
        yield server, queue
        server.close()
        await server.wait_closed()

    @pytest.fixture(name="clients")
    async def fixture_clients(self, server, port):
        client = await RpcClientCAN.connect(SHVCAN.virtualcan("test"), 42)
        server_client = await server[1].get()
        yield server_client, client
        client.disconnect()
        server_client.disconnect()
        await client.wait_disconnect()
        await server_client.wait_disconnect()

    @pytest.mark.xfail(reason="Reconnect after disconnect closes CAN bus")
    async def test_reconnect(self, clients, server):
        super().test_reconnect(clients, server)

    @pytest.mark.parametrize(
        "size",
        [*range(64), 125, 126, 127, 128, 129, 250, 251, 252, 253, 254, 255, 1039],
    )
    async def test_msg_sizes(self, clients, size):
        """Test that CAN can transmit various message sizes without mangle."""
        random.seed(42)
        msg = RpcMessage.signal(
            ".app", value=bytes(random.getrandbits(8) for _ in range(size))
        )
        await clients[0].send(msg)
        assert await clients[1].receive() == msg

    async def test_dynamic_address_single(self) -> None:
        """Check that we can assign all dynamic addresses."""
        dynamic = [await SHVCAN.virtualcan("test").dynamic_local() for _ in range(128)]
        assert sorted(local.address for local in dynamic) == list(range(128, 256))

        lastone = SHVCAN.virtualcan("test")
        with pytest.raises(RuntimeError, match="No free dynamic address is available"):
            await lastone.dynamic_local()

        for local in dynamic:
            await local.deactivate()
        await lastone.terminate()

    async def test_dynamic_address_concurrent(self) -> None:
        """Check that we can assign dynamic addresses concurrently."""
        # We don't do full 128 here because it takes a lot of time to negotiate
        # all of them in parallel.
        dynamic = await asyncio.gather(
            *(SHVCAN.virtualcan("test").dynamic_local() for _ in range(100))
        )
        assert len(set(con.address for con in dynamic)) == 100

        for local in dynamic:
            await local.deactivate()



"""The SHV RPC type map."""

from __future__ import annotations

import collections.abc
import typing

from .. import SHVType
from .any import rpctype_any
from .base import RpcType


class RpcTypeMap(RpcType):
    """The Map type representation."""

    __obj: RpcTypeMap | None = None

    def __new__(cls, tp: RpcType = rpctype_any) -> RpcTypeMap:  # noqa: D102
        if tp is rpctype_any:
            if cls.__obj is None:
                cls.__obj = super().__new__(cls)
            return cls.__obj
        return super().__new__(cls)

    def __init__(self, tp: RpcType = rpctype_any) -> None:
        self._tp = tp

    @property
    def type(self) -> RpcType:
        """Content type allowed to be used in the Map as values."""
        return self._tp

    def __eq__(self, other: object) -> bool:
        return isinstance(other, RpcTypeMap) and self._tp == other._tp

    def __str__(self) -> str:
        return f"{{{self._tp}}}"

    def validate(  # noqa: D102
        self, value: SHVType
    ) -> typing.TypeGuard[collections.abc.Mapping[str, SHVType]]:
        return isinstance(value, collections.abc.Mapping) and all(
            isinstance(k, str) and self._tp.validate(v) for k, v in value.items()
        )


rpctype_map = RpcTypeMap()
"""The singleton for :class:`RpcTypeMap`."""



"""Connection over serial device."""

import asyncio
import collections.abc
import contextlib
import fcntl
import logging
import os
import pathlib
import termios
import tty
import typing

from .abc import RpcClient, RpcServer
from .stream import RpcClientStream, RpcProtocolSerialCRC

try:
    import asyncinotify
except (ImportError, TypeError):  # pragma: no cover
    asyncinotify = None

logger = logging.getLogger(__name__)


# TODO this is not supported on windows
class RpcClientTTY(RpcClientStream):
    """RPC connection to some SHV peer over serial communication device."""

    def __init__(
        self,
        port: str,
        baudrate: int = 115200,
    ) -> None:
        super().__init__(RpcProtocolSerialCRC)
        self.port = port
        try:
            self._bspeed = getattr(termios, f"B{baudrate}")
        except AttributeError as exc:
            raise ValueError(f"Unsupported baudrate {baudrate}") from exc

    def __str__(self) -> str:
        return f"tty:{self.port}"

    async def reset(self) -> None:  # noqa: D102
        was_connected = self.connected
        await super().reset()
        if not was_connected and self.connected:
            await self.reset()  # Nested reset to send reset message

    async def _open_connection(
        self,
    ) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        loop = asyncio.get_running_loop()

        fd = os.open(self.port, os.O_RDWR)
        try:
            if not os.isatty(fd):
                raise ValueError(f"{self.port} is not TTY")
            fcntl.flock(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
            attr = termios.tcgetattr(fd)
            _cfmakeraw(attr)
            attr[tty.CFLAG] |= termios.CRTSCTS  # flow control
            attr[tty.ISPEED] = attr[tty.OSPEED] = self._bspeed
            termios.tcsetattr(fd, termios.TCSAFLUSH, attr)
        except Exception:
            os.close(fd)
            raise
        file = os.fdopen(fd, "wb")

        reader = asyncio.StreamReader()
        rprotocol = asyncio.StreamReaderProtocol(reader)
        await loop.connect_read_pipe(lambda: rprotocol, file)

        wprotocol = asyncio.StreamReaderProtocol(asyncio.StreamReader())
        wtransport, _ = await loop.connect_write_pipe(lambda: wprotocol, file)
        writer = asyncio.StreamWriter(wtransport, wprotocol, None, loop)

        return reader, writer


class RpcServerTTY(RpcServer):
    """RPC server waiting for TTY to appear.

    This actually only maintains a single client as there can't be more than one
    client on single TTY.
    """

    def __init__(
        self,
        client_connected_cb: collections.abc.Callable[
            [RpcClient], collections.abc.Awaitable[None] | None
        ],
        port: str,
        baudrate: int = 115200,
    ) -> None:
        self.client_connected_cb = client_connected_cb
        """Callbact that is called when new client is connected."""
        self.client = RpcClientTTY(port, baudrate)
        """The :class:`RpcClientTTY` instance."""
        self._task: asyncio.Task | None = None

    def __str__(self) -> str:
        return f"server.tty:{self.client.port}"

    async def _loop(self) -> None:
        while True:
            try:
                await self.client.reset()
            except OSError as exc:
                logger.debug("%s: Waiting for accessible TTY device: %s", self, exc)
            else:
                logger.debug("%s: Openned", self)
                res = self.client_connected_cb(self.client)
                if isinstance(res, collections.abc.Awaitable):
                    await res
                await self.client.wait_disconnect()
                continue
            if asyncinotify is not None:
                with asyncinotify.Inotify() as inotify:
                    pth = pathlib.Path(self.client.port)
                    inotify.add_watch(
                        pth.parent, asyncinotify.Mask.CREATE | asyncinotify.Mask.ATTRIB
                    )
                    async for event in inotify:
                        if str(pth.name) == str(event.name):
                            break
            else:
                await asyncio.sleep(5)

    def is_serving(self) -> bool:  # noqa: D102
        return self._task is not None and not self._task.done()

    async def listen(self) -> None:  # noqa: D102
        if not self.is_serving():
            self._task = asyncio.create_task(self._loop())

    async def listen_forever(self) -> None:  # noqa: D102
        await self.listen()
        await self.wait_closed()

    def close(self) -> None:  # noqa: D102
        if self._task is not None:
            self._task.cancel()

    async def wait_closed(self) -> None:  # noqa: D102
        if self._task is not None:
            with contextlib.suppress(asyncio.CancelledError):
                await self._task

    def terminate(self) -> None:  # noqa D102
        self.close()
        self.client.disconnect()

    async def wait_terminated(self) -> None:  # noqa D102
        await self.wait_closed()
        await self.client.wait_disconnect()


def _cfmakeraw(mode: typing.Any) -> None:  # noqa: ANN401
    """Make termios mode raw.

    License: Python Software Foundation License Version 2
    Source: CPython (tty.py)
    Author: Steen Lumholt.

    This is copy of the code provided in standard Python library since Python
    3.12. The switch to the ``tty.cfmakeraw`` should be made once minimal
    version is at least that.
    """
    # Clear all POSIX.1-2017 input mode flags.
    # See chapter 11 "General Terminal Interface"
    # of POSIX.1-2017 Base Definitions.
    mode[tty.IFLAG] &= ~(
        termios.IGNBRK
        | termios.BRKINT
        | termios.IGNPAR
        | termios.PARMRK
        | termios.INPCK
        | termios.ISTRIP
        | termios.INLCR
        | termios.IGNCR
        | termios.ICRNL
        | termios.IXON
        | termios.IXANY
        | termios.IXOFF
    )

    # Do not post-process output.
    mode[tty.OFLAG] &= ~termios.OPOST

    # Disable parity generation and detection; clear character size mask;
    # let character size be 8 bits.
    mode[tty.CFLAG] &= ~(termios.PARENB | termios.CSIZE)
    mode[tty.CFLAG] |= termios.CS8

    # Clear all POSIX.1-2017 local mode flags.
    mode[tty.LFLAG] &= ~(
        termios.ECHO
        | termios.ECHOE
        | termios.ECHOK
        | termios.ECHONL
        | termios.ICANON
        | termios.IEXTEN
        | termios.ISIG
        | termios.NOFLSH
        | termios.TOSTOP
    )

    # POSIX.1-2017, 11.1.7 Non-Canonical Mode Input Processing,
    # Case B: MIN>0, TIME=0
    # A pending read shall block until MIN (here 1) bytes are received,
    # or a signal is received.
    mode[tty.CC] = list(mode[tty.CC])
    mode[tty.CC][termios.VMIN] = 1
    mode[tty.CC][termios.VTIME] = 0



"""Implementation of Rpc RPC specific errors."""

from __future__ import annotations

import contextlib
import enum
import traceback
import typing

from ..value import SHVType, is_shvimap


class RpcErrorCode(enum.IntEnum):
    """Number representing an SHV RPC error code."""

    NO_ERROR = 0
    METHOD_NOT_FOUND = 2
    INVALID_PARAM = 3
    METHOD_CALL_EXCEPTION = 8
    UNKNOWN = 9
    LOGIN_REQUIRED = 10
    USER_ID_REQUIRED = 11
    NOT_IMPLEMENTED = 12
    TRY_AGAIN_LATER = 13
    REQUEST_INVALID = 14
    USER_CODE = 32


class RpcError(RuntimeError):
    """Top level for Rpc RPC errors.

    This tries to be a bit smart when objects are created. The
    :attr:`shv_error_map` is used to lookup a more appropriate error object.
    If you are adding any of your own error codes you can add them to this map
    to be looked up by their code.

    :param msg: Message describing the error circumstances.
    :param code: Error code.
    """

    class Key(enum.IntEnum):
        """Keys in the error IMap."""

        CODE = 1
        MESSAGE = 2

    shv_error_code: typing.ClassVar[RpcErrorCode] = RpcErrorCode.UNKNOWN
    shv_error_map: typing.Final[dict[int, type[RpcError]]] = {}

    def __new__(
        cls, msg: str | None = None, code: RpcErrorCode | None = None
    ) -> RpcError:
        """Create an appropriate exception based on the code."""
        ncls = cls.shv_error_map.get(cls.shv_error_code if code is None else code, cls)
        return super(RpcError, cls).__new__(ncls)  # noqa UP008

    def __init__(
        self, msg: str | None = None, code: RpcErrorCode | None = None
    ) -> None:
        super().__init__(msg, self.shv_error_code if code is None else code)

    def __init_subclass__(cls, *args: typing.Any, **kwargs: typing.Any) -> None:  # noqa: ANN401
        super().__init_subclass__(*args, **kwargs)
        if cls.shv_error_code in RpcError.shv_error_map:
            raise TypeError(f"RPC Error already defined for {cls.shv_error_code}")
        RpcError.shv_error_map[cls.shv_error_code] = cls

    @property
    def message(self) -> str | None:
        """Provides access to the SHV RPC message."""
        return typing.cast(str | None, self.args[0])

    @property
    def error_code(self) -> RpcErrorCode:
        """Provides access to the :class:`shv.rpcdef.errors.RpcErrorCode`."""
        return RpcErrorCode(self.args[1])

    def to_shv(self) -> SHVType:
        """Convert to SHV RPC representation."""
        res: dict[int, SHVType] = {
            self.Key.CODE: self.error_code,
            **({} if self.message is None else {self.Key.MESSAGE: self.message}),
        }
        return res

    @classmethod
    def from_shv(cls, value: SHVType) -> RpcError:
        """Create from SHV RPC representation."""
        if not is_shvimap(value):
            raise ValueError(f"Expected IMap but got {value!r}.")
        msg = value.get(cls.Key.MESSAGE)
        if not isinstance(msg, str | None):
            raise ValueError(f"Invalid Message format: {msg}")
        rcode = value.get(cls.Key.CODE)
        if not isinstance(rcode, int):
            raise ValueError(f"Invalid Code format: {rcode}")
        code = RpcErrorCode.UNKNOWN
        with contextlib.suppress(ValueError):
            code = RpcErrorCode(rcode)
        return cls(msg, code)

    @staticmethod
    def from_exception(exc: Exception) -> RpcError:
        """Create exception from other Python exception."""
        if isinstance(exc, RpcError):
            return exc
        return RpcMethodCallExceptionError("".join(traceback.format_exception(exc)))


class RpcMethodNotFoundError(RpcError):
    """The method does not exist or is not available."""

    shv_error_code = RpcErrorCode.METHOD_NOT_FOUND


class RpcInvalidParamError(RpcError):
    """Invalid method parameter were provided."""

    shv_error_code = RpcErrorCode.INVALID_PARAM


class RpcMethodCallExceptionError(RpcError):
    """Method call resulted in exception and not result."""

    shv_error_code = RpcErrorCode.METHOD_CALL_EXCEPTION


class RpcLoginRequiredError(RpcError):
    """Login sequence must be performed before anything else."""

    shv_error_code = RpcErrorCode.LOGIN_REQUIRED


class RpcUserIDRequiredError(RpcError):
    """Request must be sent with UserID field."""

    shv_error_code = RpcErrorCode.USER_ID_REQUIRED


class RpcNotImplementedError(RpcError):
    """Called method that is not implemented right now but valid."""

    shv_error_code = RpcErrorCode.NOT_IMPLEMENTED


class RpcTryAgainLaterError(RpcError):
    """The resource to fullfill request are temporally unavailable."""

    shv_error_code = RpcErrorCode.TRY_AGAIN_LATER


class RpcRequestInvalidError(RpcError):
    """There is no such known request (in response to abort request)."""

    shv_error_code = RpcErrorCode.REQUEST_INVALID



"""The definition of standard SHV types."""

from __future__ import annotations

import typing

from .. import SHVType
from .any import rpctype_any
from .base import RpcType
from .bitfield import RpcTypeBitfield
from .blob import rpctype_blob
from .bool import rpctype_bool
from .datetime import rpctype_datetime
from .enum import RpcTypeEnum
from .integer import RpcTypeInteger, rpctype_integer
from .list import RpcTypeList
from .map import RpcTypeMap, rpctype_map
from .oneof import RpcTypeOptional
from .string import rpctype_string
from .struct import RpcTypeStruct
from .unsigned import rpctype_unsigned


class RpcTypeStandard(RpcType):
    """The container for the types defined by standard."""

    def __init__(self, name: str, tp: RpcType) -> None:
        self._name = name
        self._tp = tp

    @property
    def name(self) -> str:
        """Name assigned to the standard type."""
        return self._name

    @property
    def type(self) -> RpcType:
        """Type this standard type represents."""
        return self._tp

    def __str__(self) -> str:
        return f"!{self._name}"

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, RpcTypeStandard)
            and self._name == other._name
            and self._tp == other._tp
        )

    def validate(self, value: SHVType) -> typing.TypeGuard[SHVType]:  # noqa: D102
        return self._tp.validate(value)


rpctype_dir = RpcTypeStandard(
    "dir",
    RpcTypeStruct({
        1: (rpctype_string, "name"),
        2: (
            RpcTypeOptional(
                RpcTypeBitfield(
                    (1, rpctype_bool, "isGetter"),
                    (2, rpctype_bool, "isSetter"),
                    (3, rpctype_bool, "largeResult"),
                    (4, rpctype_bool, "notIndempotent"),
                    (5, rpctype_bool, "userIDRequired"),
                    (6, rpctype_bool, "isUpdatable"),
                )
            ),
            "flags",
        ),
        3: (RpcTypeOptional(rpctype_string), "paramType"),
        4: (RpcTypeOptional(rpctype_string), "resultType"),
        5: (RpcTypeInteger(0, 63), "accessLevel"),
        6: (RpcTypeMap(RpcTypeOptional(rpctype_string)), "signals"),
        63: (rpctype_map, "extra"),
    }),
)
"""
i{s:name:1,u[b:isGetter:1,b:isSetter,b:largeResult,b:notIndempotent,b:userIDRequired,b:isUpdatable]|n:flags,s|n:paramType,s|n:resultType,i(0,63):accessLevel,{s|n}:signals,{?}:extra:63}
"""


rpctype_alert = RpcTypeStandard(
    "alert",
    RpcTypeStruct({
        0: (rpctype_datetime, "date"),
        1: (RpcTypeInteger(0, 63), "level"),
        2: (rpctype_string, "id"),
        3: (rpctype_any, "info"),
    }),
)
"""
i{t:date,i(0,63):level,s:id,?:info}
"""

rpctype_clientinfo = RpcTypeStandard(
    "clientInfo",
    RpcTypeStruct({
        1: (rpctype_integer, "clientId"),
        2: (RpcTypeOptional(rpctype_string), "userName"),
        3: (RpcTypeOptional(rpctype_string), "mountPoint"),
        4: (
            RpcTypeOptional(RpcTypeMap(RpcTypeOptional(rpctype_integer))),
            "subscriptions",
        ),
        63: (rpctype_map, "extra"),
    }),
)
"""
i{i:clientId:1,s|n:userName,s|n:mountPoint,{i|n}|n:subscriptions,{?}:extra:63}
"""


rpctype_stat = RpcTypeStandard(
    "stat",
    RpcTypeStruct({
        0: (rpctype_integer, "type"),
        1: (rpctype_integer, "size"),
        2: (rpctype_integer, "pageSize"),
        3: (RpcTypeOptional(rpctype_datetime), "accessTime"),
        4: (RpcTypeOptional(rpctype_datetime), "modTime"),
        5: (RpcTypeOptional(rpctype_integer), "maxWrite"),
    }),
)
"""
i{i:type,i:size,i:pageSize,t|n:accessTime,t|n:modTime,i|n:maxWrite}
"""

rpctype_exchange_p = RpcTypeStandard(
    "exchangeP",
    RpcTypeStruct({
        0: (rpctype_unsigned, "counter"),
        1: (RpcTypeOptional(rpctype_unsigned), "readyToReceive"),
        3: (RpcTypeOptional(rpctype_blob), "data"),
    }),
)
"""
i{u:counter,u|n:readyToReceive,b|n:data:3}
"""

rpctype_exchange_r = RpcTypeStandard(
    "exchangeR",
    RpcTypeStruct({
        1: (RpcTypeOptional(rpctype_unsigned), "readyToReceive"),
        2: (RpcTypeOptional(rpctype_unsigned), "readyToSend"),
        3: (RpcTypeOptional(rpctype_blob), "data"),
    }),
)
"""
i{u|n:readyToReceive:1,u|n:readyToSend,b|n:data}
"""

rpctype_exchange_v = RpcTypeStandard(
    "exchangeV",
    RpcTypeStruct({
        1: (RpcTypeOptional(rpctype_unsigned), "readyToReceive"),
        2: (RpcTypeOptional(rpctype_unsigned), "readyToSend"),
    }),
)
"""
i{u|n:readyToReceive:1,u|n:readyToSend}
"""

rpctype_getlog_p = RpcTypeStandard(
    "getLogP",
    RpcTypeStruct({
        1: (RpcTypeOptional(rpctype_datetime), "since"),
        2: (RpcTypeOptional(rpctype_datetime), "until"),
        3: (RpcTypeOptional(RpcTypeInteger(0)), "count"),
        4: (RpcTypeOptional(rpctype_string), "ri"),
    }),
)
"""
i{t|n:since:1,t|n:until,i(0,)|n:count,s|n:ri}
"""

rpctype_getlog_r = RpcTypeStandard(
    "getLogR",
    RpcTypeList(
        RpcTypeStruct({
            1: (RpcTypeOptional(rpctype_datetime), "timestamp"),
            2: (RpcTypeOptional(RpcTypeInteger(0)), "ref"),
            3: (RpcTypeOptional(rpctype_string), "path"),
            4: (RpcTypeOptional(rpctype_string), "signal"),
            5: (RpcTypeOptional(rpctype_string), "source"),
            6: (rpctype_any, "value"),
            7: (RpcTypeOptional(rpctype_string), "userId"),
            8: (RpcTypeOptional(rpctype_bool), "repeat"),
        })
    ),
)
"""
[i{t|n:timestamp:1,i(0,)|n:ref,s|n:path,s|n:signal,s|n:source,?:value,s|n:userId,b|n:repeat}]
"""


rpctype_history_records = RpcTypeStandard(
    "historyRecords",
    RpcTypeList(
        RpcTypeStruct({
            0: (
                RpcTypeEnum({1: "normal", 2: "keep", 3: "timeJump", 4: "timeAbig"}),
                "type",
            ),
            1: (rpctype_datetime, "timestamp"),
            2: (RpcTypeOptional(rpctype_string), "path"),
            3: (RpcTypeOptional(rpctype_string), "signal"),
            4: (RpcTypeOptional(rpctype_string), "source"),
            5: (rpctype_any, "value"),
            6: (RpcTypeInteger(0, 63), "accessLevel"),
            7: (RpcTypeOptional(rpctype_string), "userId"),
            8: (RpcTypeOptional(rpctype_bool), "repeat"),
            60: (RpcTypeOptional(rpctype_integer), "timeJump"),
        })
    ),
)
"""
[i{i[normal:1,keep,timeJump,timeAbig]:type,t:timestamp,s|n:path,s|n:signal,s|n:source,?:value,i:accessLevel,s|n:userId,b|n:repeat,i:timeJump:60}]
"""



"""Check the SHV path correctness."""

import os

import pytest

from shv.path import SHVPath


def test_nospace():
    with pytest.raises(ValueError):
        _ = SHVPath("some", "with path")


@pytest.mark.parametrize(
    "path,expected",
    (
        (SHVPath(), tuple()),
        (SHVPath("/some"), ("some",)),
        (SHVPath("/some/extra/node"), ("some", "extra", "node")),
    ),
)
def test_parts(path, expected):
    assert path.parts == expected


@pytest.mark.parametrize(
    "path,expected",
    (
        (SHVPath(), ""),
        (SHVPath("/some"), "some"),
        (SHVPath("/some", ".."), ""),
        (SHVPath("", ".."), ""),
        (SHVPath("some", ""), "some"),
        (SHVPath("some/node", "/from/root"), "from/root"),
    ),
)
def test_str(path, expected):
    assert str(path) == expected


def test_osfspath():
    assert os.fspath(SHVPath("test/node")) == "test/node"


def test_repr():
    assert repr(SHVPath("test/node")) == "SHVPath('test/node')"


@pytest.mark.parametrize(
    "path,expected",
    ((SHVPath(), False), (SHVPath("node"), True)),
)
def test_bool(path, expected):
    assert bool(path) is expected


def test_strpath():
    assert (SHVPath("some") / "node") == "some/node"


def test_pathstr():
    assert ("some" / SHVPath("node")) == SHVPath("some/node")


def test_hash():
    path = SHVPath()
    dct = {path: True, SHVPath("some/node"): False}
    assert dct[path] is True
    assert dct[SHVPath()] is True  # Intentional unused of path
    assert dct[SHVPath("some/node")] is False
    assert dct.get(SHVPath("some")) is None
    assert dct.get(tuple()) is None


def test_cmp():
    assert SHVPath() < "foo"
    assert SHVPath("alice") < SHVPath("foo")
    assert not SHVPath("foo") < "alice"
    assert SHVPath("foo") <= SHVPath("foo")
    assert SHVPath("foo") > SHVPath("alice")
    assert SHVPath("foo") >= SHVPath("foo")

    with pytest.raises(TypeError):
        assert SHVPath() < 1
    with pytest.raises(TypeError):
        assert SHVPath() <= 1
    with pytest.raises(TypeError):
        assert SHVPath() > 1
    with pytest.raises(TypeError):
        assert SHVPath() >= 1


def test_parents():
    path = SHVPath("some/node/here/after")
    assert list(path.parents) == [
        SHVPath("some/node/here"),
        SHVPath("some/node"),
        SHVPath("some"),
        SHVPath(""),
    ]
    assert SHVPath("some/node").parents[-1] == SHVPath("")
    assert SHVPath("some/node").parents[-2] == SHVPath("some")
    with pytest.raises(IndexError):
        SHVPath("some/node").parents[3]
    with pytest.raises(IndexError):
        SHVPath("some/node").parents[-3]
    assert path.parents[1:4:2] == (SHVPath("some/node"), SHVPath())


def test_parent():
    assert SHVPath("some/node/here").parent == "some/node"


def test_root_parent():
    """Root is its own parent."""
    root = SHVPath("")
    assert root is root.parent


def test_name():
    assert SHVPath("some/node/here").name == "here"


def test_suffix():
    assert SHVPath("some/node/here.txt").suffix == ".txt"


def test_suffixes():
    assert SHVPath("some/node/here.txt.tar.gz").suffixes == [".txt", ".tar", ".gz"]


def test_stem():
    assert SHVPath("some/node/here.txt.tar.gz").stem == "here"


def test_is_relative_to():
    assert SHVPath().is_relative_to("")
    assert SHVPath("some/node").is_relative_to("")
    assert SHVPath("some/node").is_relative_to("some")
    assert SHVPath("some/node").is_relative_to(SHVPath("some/node"))
    assert not SHVPath("some/node").is_relative_to("som")
    assert not SHVPath("some/node").is_relative_to("some/n")


def test_joinpath():
    assert SHVPath().joinpath("some", "node/here") == "some/node/here"


def test_full_match():
    assert SHVPath("some/node/here").full_match("some/**")
    assert not SHVPath("some/node/here").full_match("some/*")
    assert not SHVPath("some/node/here").full_match("*/here")


def test_match():
    assert SHVPath("some/node/here").match("node/*")
    assert not SHVPath("some/node/here").match("here/*")


def test_relative_to():
    assert SHVPath("some/node").relative_to("some") == "node"
    with pytest.raises(ValueError):
        assert SHVPath("some/node").relative_to(SHVPath("other"))


def test_with_name():
    assert SHVPath("some/node").with_name("other") == "some/other"
    with pytest.raises(ValueError):
        SHVPath().with_name("node")


def test_with_stem():
    assert SHVPath("some/node.txt.gz").with_stem("other") == "some/other.txt.gz"
    with pytest.raises(ValueError):
        SHVPath().with_stem("node")


def test_with_suffix():
    assert SHVPath("some/node.txt.gz").with_suffix(".xz") == "some/node.xz"
    with pytest.raises(ValueError):
        SHVPath().with_suffix(".xz")



"""SHV RPC definitions."""

from .access import RpcAccess
from .dir import RpcDir
from .errors import (
    RpcError,
    RpcInvalidParamError,
    RpcLoginRequiredError,
    RpcMethodCallExceptionError,
    RpcMethodNotFoundError,
    RpcNotImplementedError,
    RpcRequestInvalidError,
    RpcTryAgainLaterError,
    RpcUserIDRequiredError,
)

__all__ = [
    "RpcAccess",
    "RpcDir",
    "RpcError",
    "RpcInvalidParamError",
    "RpcLoginRequiredError",
    "RpcMethodCallExceptionError",
    "RpcMethodNotFoundError",
    "RpcNotImplementedError",
    "RpcRequestInvalidError",
    "RpcTryAgainLaterError",
    "RpcUserIDRequiredError",
]



"""Python implementation of Silicon Heaven."""

from .__version__ import VERSION
from .shvversion import SHV_VERSION, SHV_VERSION_MAJOR, SHV_VERSION_MINOR
from .value import (
    SHVBool,
    SHVBoolType,
    SHVBytes,
    SHVDatetime,
    SHVDecimal,
    SHVFloat,
    SHVIMap,
    SHVIMapType,
    SHVInt,
    SHVList,
    SHVListType,
    SHVMap,
    SHVMapType,
    SHVMeta,
    SHVMetaType,
    SHVNull,
    SHVNullType,
    SHVStr,
    SHVType,
    SHVUInt,
    decimal_rexp,
    is_shvbool,
    is_shvimap,
    is_shvlist,
    is_shvmap,
    is_shvnull,
    is_shvtype,
    shvmeta,
    shvmeta_eq,
)

__all__ = [
    "SHV_VERSION",
    "SHV_VERSION_MAJOR",
    "SHV_VERSION_MINOR",
    "VERSION",
    "SHVBool",
    "SHVBoolType",
    "SHVBytes",
    "SHVDatetime",
    "SHVDecimal",
    "SHVFloat",
    "SHVIMap",
    "SHVIMapType",
    "SHVInt",
    "SHVList",
    "SHVListType",
    "SHVMap",
    "SHVMapType",
    "SHVMeta",
    "SHVMetaType",
    "SHVNull",
    "SHVNullType",
    "SHVStr",
    "SHVType",
    "SHVUInt",
    "decimal_rexp",
    "is_shvbool",
    "is_shvimap",
    "is_shvlist",
    "is_shvmap",
    "is_shvnull",
    "is_shvtype",
    "shvmeta",
    "shvmeta_eq",
]



"""Chainpack data format reader and writer."""

from __future__ import annotations

import datetime
import decimal
import enum
import io
import struct
import typing

from . import commonpack
from .value import (
    SHVIMap,
    SHVIMapType,
    SHVListType,
    SHVMap,
    SHVMapType,
    SHVMeta,
    SHVMetaType,
    SHVType,
    SHVUInt,
    decimal_rexp,
)


class ChainPack:
    """ChainPack constans and definitions."""

    ProtocolType = 1

    class Schema(enum.IntEnum):
        """The packing schema type association."""

        CP_Null = 128
        CP_UInt = 129
        CP_Int = 130
        CP_Double = 131
        CP_Bool = 132
        CP_Blob = 133
        CP_String = 134  # utf8 encoded string
        CP_List = 136
        CP_Map = 137
        CP_IMap = 138
        CP_MetaMap = 139
        CP_Decimal = 140
        CP_DateTime = 141
        CP_CString = 142
        CP_FALSE = 253
        CP_TRUE = 254
        CP_TERM = 255

    # UTC msec since 2.2. 2018
    # Fri Feb 02 2018 00:00:00 == 1517529600 EPOCH
    SHV_EPOCH_SEC = 1517529600
    # ChainPack.INVALID_MIN_OFFSET_FROM_UTC = (-64 * 15)

    @staticmethod
    def unpack(data: bytes | str) -> SHVType:
        """Unpack single value from given data."""
        return ChainPackReader.unpack(data)

    @staticmethod
    def pack(value: SHVType) -> bytes:
        """Pack given value and return bytes."""
        return ChainPackWriter.pack(value)

    @staticmethod
    def unpack_uint_data(data: bytes | bytearray) -> int:
        """Unpack given value as unsigned int data.

        The existence of this is pretty much just to support Stream transport
        protocol for SHV RPC. You probably do not want to use this for anything
        else.

        :param data: Data that should contain unsigned integer in Chainpack
          format. Unused bytes at the end are ignored.
        :return: unpacked integer.
        :raise ValueError: in case of invalid data.
        """
        try:
            return ChainPackReader(data).read_uint_data()
        except EOFError as exc:
            raise ValueError from exc

    @staticmethod
    def pack_uint_data(value: int) -> bytes:
        """Pack given value as unsigned int data.

        The existence of this is pretty much just to support Stream transport
        protocol for SHV RPC. You probably do not want to use this for anything
        else.

        :param value: Unsigned integer to be packed.
        :return: bytes with unsigned integer.
        """
        bio = io.BytesIO()
        writer = ChainPackWriter(bio)
        writer.write_uint_data(value)
        return bio.getvalue()


class ChainPackReader(commonpack.CommonReader):
    """Read data in ChainPack format."""

    def read(self) -> SHVType:  # noqa: D102
        packing_schema = self._read_byte()
        if packing_schema < 64:
            return SHVUInt(packing_schema & 63)
        if packing_schema < 128:
            return packing_schema & 63
        if packing_schema == ChainPack.Schema.CP_Null:
            return None
        if packing_schema == ChainPack.Schema.CP_TRUE:
            return True
        if packing_schema == ChainPack.Schema.CP_FALSE:
            return False
        if packing_schema == ChainPack.Schema.CP_Int:
            return self._read_int_data()
        if packing_schema == ChainPack.Schema.CP_UInt:
            return SHVUInt(self.read_uint_data())
        if packing_schema == ChainPack.Schema.CP_Double:
            return self._read_double()
        if packing_schema == ChainPack.Schema.CP_Decimal:
            return self._read_decimal()
        if packing_schema == ChainPack.Schema.CP_DateTime:
            return self._read_datetime()
        if packing_schema == ChainPack.Schema.CP_Map:
            if mvalue := typing.cast(SHVMapType, self._read_map()):
                return mvalue
            return SHVMap()  # Remove confusin between map and imap
        if packing_schema == ChainPack.Schema.CP_IMap:
            if imvalue := typing.cast(SHVIMapType, self._read_map()):
                return imvalue
            return SHVIMap()  # Remove confusin between map and imap
        if packing_schema == ChainPack.Schema.CP_List:
            return self._read_list()
        if packing_schema == ChainPack.Schema.CP_Blob:
            return self._read_blob()
        if packing_schema == ChainPack.Schema.CP_String:
            return self._read_string()
        if packing_schema == ChainPack.Schema.CP_CString:
            return self._read_cstring()
        if packing_schema == ChainPack.Schema.CP_MetaMap:
            meta = self._read_map()
            return SHVMeta.new(self.read(), meta)
        raise ValueError(f"ChainPack - Invalid type: {packing_schema}")

    def _read_uint_data_helper(self) -> tuple[int, int]:
        num = 0
        bitlen = 0
        head = self._read_byte()
        if (head & 128) == 0:
            bytes_to_read_cnt = 0
            num = head & 127
            bitlen = 7
        elif (head & 64) == 0:
            bytes_to_read_cnt = 1
            num = head & 63
            bitlen = 6 + 8
        elif (head & 32) == 0:
            bytes_to_read_cnt = 2
            num = head & 31
            bitlen = 5 + 2 * 8
        elif (head & 16) == 0:
            bytes_to_read_cnt = 3
            num = head & 15
            bitlen = 4 + 3 * 8
        else:
            bytes_to_read_cnt = (head & 0xF) + 4
            bitlen = bytes_to_read_cnt * 8

        for _ in range(bytes_to_read_cnt):
            r = self._read_byte()
            num = (num << 8) + r
        return num, bitlen

    def read_uint_data(self) -> int:  # noqa: D102
        num, _ = self._read_uint_data_helper()
        return num

    def _read_int_data(self) -> int:
        num, bitlen = self._read_uint_data_helper()
        sign_bit_mask = 1 << (bitlen - 1)
        neg = num & sign_bit_mask
        snum = num
        if neg:
            snum &= ~sign_bit_mask
            snum = -snum
        return snum

    def _read_double(self) -> float:
        res = struct.unpack("<d", self._read(8))  # little endian
        return typing.cast(float, res[0])

    def _read_decimal(self) -> decimal.Decimal:
        mant = self._read_int_data()
        exp = self._read_int_data()
        return decimal.Decimal(f"{mant}e{exp}")

    def _read_datetime(self) -> datetime.datetime:
        d = self._read_int_data()
        offset = 0
        has_tz_offset = d & 1
        has_not_msec = d & 2
        d >>= 2
        if has_tz_offset:
            offset = d & 0x7F
            if offset >= 128:
                offset -= 128  # sign extension
            d >>= 7
        f: float = d if has_not_msec else d / 1000
        f += ChainPack.SHV_EPOCH_SEC
        tzone = datetime.timezone(datetime.timedelta(minutes=offset * 15))
        return datetime.datetime.fromtimestamp(f, tzone)

    def _read_blob(self) -> bytes:
        dlen = self.read_uint_data()
        return self._read(dlen) if dlen else b""

    def _read_string(self) -> str:
        slen = self.read_uint_data()
        if slen <= 0:
            return ""
        return self._read(slen).decode("utf-8")

    def _read_cstring(self) -> str:
        res = ""
        while True:
            b = self._read_byte()
            if b == ord("\\"):
                b = self._read_byte()
                if b == ord("\\"):
                    res += "\\"
                elif b == ord("0"):
                    res += "\0"
                else:
                    res += chr(b)
            else:
                if b == 0:
                    break  # end of string
                res += chr(b)
        return res

    def _read_list(self) -> SHVListType:
        lst = []
        while self._peek_byte() != ChainPack.Schema.CP_TERM:
            lst.append(self.read())
        self._peek_drop()
        return lst

    def _read_map(self) -> dict[str | int, SHVType]:
        mmap: dict[str | int, SHVType] = {}
        while self._peek_byte() != ChainPack.Schema.CP_TERM:
            key = self.read()
            if not isinstance(key, str | int) or isinstance(key, SHVUInt):
                raise ValueError(f"Invalid Map key: {type(key)}")
            val = self.read()
            mmap[key] = val
        self._peek_drop()
        return mmap


class ChainPackWriter(commonpack.CommonWriter):
    """Write data in ChainPack format."""

    # Integer encoding:
    #  0 ...  7 bits  1  byte  |0|s|x|x|x|x|x|x|<-- LSB
    #  8 ... 14 bits  2  bytes |1|0|s|x|x|x|x|x| |x|x|x|x|x|x|x|x|<-- LSB
    # 15 ... 21 bits  3  bytes |1|1|0|s|x|x|x|x| |x|x|x|x|x|x|x|x| |x|x|x|x|x|x|x|x|<-- LSB
    # 22 ... 28 bits  4  bytes |1|1|1|0|s|x|x|x| |x|x|x|x|x|x|x|x| |x|x|x|x|x|x|x|x| |x|x|x|x|x|x|x|x|<-- LSB
    # 29+       bits  5+ bytes |1|1|1|1|n|n|n|n| |s|x|x|x|x|x|x|x| |x|x|x|x|x|x|x|x| |x|x|x|x|x|x|x|x| ... <-- LSB
    #                                         n ==  0 ->  4 bytes number (32 bit number)
    #                                         n ==  1 ->  5 bytes number
    #                                         n == 14 -> 18 bytes number
    #                                         n == 15 -> for future (number of bytes will be specified in next byte)

    @classmethod
    def _bytes_needed(cls, bit_len: int) -> int:
        """Calculate needed maximum number of bytes.

        Return max bit length >= bit_len, which can be encoded by same number of
        bytes number of bytes needed to encode bit_len.
        """
        if bit_len <= 28:
            cnt = ((bit_len - 1) // 7) + 1
        else:
            cnt = ((bit_len - 1) // 8) + 2
        return cnt or 1  # Always at least one byte is needed

    @classmethod
    def _expand_bit_len(cls, bit_len: int) -> int:
        byte_cnt = cls._bytes_needed(bit_len)
        if bit_len <= 28:
            ret = byte_cnt * (8 - 1) - 1
        else:
            ret = (byte_cnt - 1) * 8 - 1
        return ret

    def _write_uint_data_helper(self, num: int, bit_len: int) -> None:
        byte_cnt = self._bytes_needed(bit_len)
        data = bytearray(byte_cnt)
        for i in range(byte_cnt - 1, -1, -1):
            data[i] = num & 0xFF
            num >>= 8

        if bit_len <= 28:
            mask = 0xF0 << (4 - byte_cnt)
            data[0] &= ~mask
            mask = (mask << 1) & 0xFF
            data[0] |= mask
        else:
            data[0] = 0xF0 | (byte_cnt - 5)

        for i in range(0, byte_cnt):
            self._write(data[i])

    def write_meta(self, meta: SHVMetaType) -> None:  # noqa: D102
        self._write(ChainPack.Schema.CP_MetaMap)
        for k, v in meta.items():
            self.write(k)
            self.write(v)
        self._write(ChainPack.Schema.CP_TERM)

    def write_null(self) -> None:  # noqa: D102
        self._write(ChainPack.Schema.CP_Null)

    def write_bool(self, value: bool) -> None:  # noqa: D102
        self._write(ChainPack.Schema.CP_TRUE if value else ChainPack.Schema.CP_FALSE)

    def write_blob(self, value: bytes | bytearray) -> None:  # noqa: D102
        self._write(ChainPack.Schema.CP_Blob)
        self.write_uint_data(len(value))
        self._write(value)

    def write_string(self, value: str) -> None:  # noqa: D102
        bstring = value.encode("utf-8")
        self._write(ChainPack.Schema.CP_String)
        self.write_uint_data(len(bstring))
        self._write(bstring)

    def write_cstring(self, value: str) -> None:  # noqa: D102
        bstring = value.encode("utf-8")
        self._write(ChainPack.Schema.CP_CString)
        self._write(bstring)
        self._write(b"\0")

    def write_uint(self, value: int) -> None:  # noqa: D102
        if value < 64:
            self._write(value % 64)
        else:
            self._write(ChainPack.Schema.CP_UInt)
            self.write_uint_data(value)

    def write_uint_data(self, value: int) -> None:  # noqa: D102
        bitcnt = value.bit_length()
        self._write_uint_data_helper(value, bitcnt)

    def write_int(self, value: int) -> None:  # noqa: D102
        if 0 <= value < 64:
            self._write((value % 64) + 64)
        else:
            self._write(ChainPack.Schema.CP_Int)
            self.write_int_data(value)

    def write_int_data(self, value: int) -> None:  # noqa: D102
        num: int = abs(value)
        neg: bool = value < 0

        bitlen = num.bit_length()
        bitlen += 1  # add sign bit
        if neg:
            sign_pos = self._expand_bit_len(bitlen)
            sign_bit_mask = 1 << sign_pos
            num |= sign_bit_mask
        self._write_uint_data_helper(num, bitlen)

    def write_double(self, value: float) -> None:  # noqa: D102
        self._write(ChainPack.Schema.CP_Double)
        self._write(struct.pack("<d", value))  # little endian

    def write_decimal(self, value: decimal.Decimal) -> None:  # noqa: D102
        mantissa, exponent = decimal_rexp(value)
        self._write(ChainPack.Schema.CP_Decimal)
        self.write_int_data(mantissa)
        self.write_int_data(exponent)

    def write_list(self, value: SHVListType) -> None:  # noqa: D102
        self._write(ChainPack.Schema.CP_List)
        for val in value:
            self.write(val)
        self._write(ChainPack.Schema.CP_TERM)

    def write_map(self, value: SHVMapType) -> None:  # noqa: D102
        self._write(ChainPack.Schema.CP_Map)
        for k, v in value.items():
            self.write(k)
            self.write(v)
        self._write(ChainPack.Schema.CP_TERM)

    def write_imap(self, value: SHVIMapType) -> None:  # noqa: D102
        self._write(ChainPack.Schema.CP_IMap)
        for k, v in value.items():
            self.write(k)
            self.write(v)
        self._write(ChainPack.Schema.CP_TERM)

    def write_datetime(self, value: datetime.datetime) -> None:  # noqa: D102
        self._write(ChainPack.Schema.CP_DateTime)
        res = int(value.timestamp() * 1000) - (ChainPack.SHV_EPOCH_SEC * 1000)
        tzdelta = value.utcoffset()
        tzoff = int(tzdelta.total_seconds() // 60 // 15) if tzdelta is not None else 0
        if not -63 <= tzoff <= 63:
            raise ValueError(f"Invalid UTC offset value: {tzoff}")
        ms = res % 1000 == 0
        if ms:
            res //= 1000
        if tzoff != 0:
            res <<= 7
            res |= tzoff & 0x7F
        res <<= 2
        if tzoff != 0:
            res |= 1
        if ms:
            res |= 2
        self.write_int_data(res)



"""SHV RPC type blob."""

from __future__ import annotations

import typing

from .. import SHVType
from .base import RpcType


class RpcTypeBlob(RpcType):
    """The Blob type representation."""

    __obj: RpcTypeBlob | None = None

    def __new__(  # noqa: D102
        cls,
        minlen: int = 0,
        maxlen: int | None = None,
    ) -> RpcTypeBlob:
        if minlen == 0 and maxlen is None:
            if cls.__obj is None:
                cls.__obj = super().__new__(cls)
            return cls.__obj
        return super().__new__(cls)

    def __init__(
        self,
        minlen: int = 0,
        maxlen: int | None = None,
    ) -> None:
        if maxlen is not None and minlen > maxlen:
            raise ValueError("Minimum is greater than maximum")
        if minlen < 0 or (maxlen is not None and maxlen < 0):
            raise ValueError("Blob size can be only positive number")
        self._min = minlen
        self._max = maxlen

    @property
    def minlen(self) -> int:
        """Minimum length of the blob."""
        return self._min

    @property
    def maxlen(self) -> int | None:
        """Maximum length of the blob."""
        return self._max

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, RpcTypeBlob)
            and self._min == other._min
            and self._max == other._max
        )

    def __str__(self) -> str:
        lim = ""
        if self._min != 0 or self._max is not None:
            if self._min == self._max:
                lim = f"({self._max})"
            else:
                lim = f"({self._min or ''},{self._max or ''})"
        return f"x{lim}"

    def validate(self, value: SHVType) -> typing.TypeGuard[bytes]:  # noqa: D102
        return (
            isinstance(value, bytes)
            and (self._min is None or len(value) >= self._min)
            and (self._max is None or len(value) <= self._max)
        )


rpctype_blob = RpcTypeBlob()
"""Singleton for :class:`RpcTypeBlob` with no blob size limitation."""



"""The SHV RPC type Null."""

from __future__ import annotations

import typing

from .. import SHVType
from ..value import is_shvnull
from .base import RpcType


class RpcTypeNull(RpcType):
    """The Null type representation."""

    __obj = None

    def __new__(cls) -> RpcTypeNull:  # noqa: D102
        if cls.__obj is None:
            cls.__obj = object.__new__(cls)
        return cls.__obj

    def __str__(self) -> str:
        return "n"

    @staticmethod
    def validate(value: SHVType) -> typing.TypeGuard[None]:  # noqa: D102
        return is_shvnull(value)


rpctype_null = RpcTypeNull()
"""The singleton for :class:`RpcTypeNull`."""



"""SHV RPC type decimal."""

from __future__ import annotations

import decimal
import typing

from .. import SHVType
from ._tools import strdec as _strdec
from ._tools import strnum as _strnum
from .base import RpcType


class RpcTypeDecimal(RpcType):
    """The Decimal type representation."""

    __obj: RpcTypeDecimal | None = None

    def __new__(  # noqa: D102
        cls,
        minimum: decimal.Decimal | None = None,
        maximum: decimal.Decimal | None = None,
        precision: int | None = None,
        unit: str = "",
    ) -> RpcTypeDecimal:
        if minimum is None and maximum is None and precision is None and not unit:
            if cls.__obj is None:
                cls.__obj = super().__new__(cls)
            return cls.__obj
        return super().__new__(cls)

    def __init__(
        self,
        minimum: decimal.Decimal | None = None,
        maximum: decimal.Decimal | None = None,
        precision: int | None = None,
        unit: str = "",
    ) -> None:
        if minimum is not None and maximum is not None and minimum > maximum:
            raise ValueError("Minimum is greater than maximum")
        self._min = minimum
        self._max = maximum
        self._precs = precision
        if any(c in "[]{}():,|" for c in unit):
            raise ValueError("Unit contains forbidden characters")
        self._unit = unit

    @property
    def minimum(self) -> decimal.Decimal | None:
        """Minimum value allowed for this Decimal."""
        return self._min

    @property
    def maximum(self) -> decimal.Decimal | None:
        """Maximum value allowed for this Decimal."""
        return self._max

    @property
    def precision(self) -> int | None:
        """Precision of this decimal."""
        return self._precs

    @property
    def unit(self) -> str:
        """Unit to be appended to the number."""
        return self._unit

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, RpcTypeDecimal)
            and self._min == other._min
            and self._max == other._max
            and self._precs == other._precs
            and self._unit == other._unit
        )

    def __str__(self) -> str:
        lim = ""
        if self._min is not None or self._max is not None or self._precs is not None:
            precs = "" if self._precs is None else f",{_strnum(self._precs)}"
            lim = f"({_strdec(self._min)},{_strdec(self._max)}{precs})"
        return f"d{lim}{self._unit}"

    def validate(self, value: SHVType) -> typing.TypeGuard[decimal.Decimal]:  # noqa: D102
        return (
            isinstance(value, decimal.Decimal)
            and (self._min is None or value >= self._min)
            and (self._max is None or value <= self._max)
            # TODO precision
        )


rpctype_decimal = RpcTypeDecimal()
"""The singleton for :class:`RpcTypeDecimal` with default parameters."""



"""SHV RPC type Double."""

from __future__ import annotations

import typing

from .. import SHVType
from .base import RpcType


class RpcTypeDouble(RpcType):
    """The Double type representation."""

    __obj: RpcTypeDouble | None = None

    def __new__(cls, unit: str = "") -> RpcTypeDouble:  # noqa: D102
        if not unit:
            if cls.__obj is None:
                cls.__obj = super().__new__(cls)
            return cls.__obj
        return super().__new__(cls)

    def __init__(self, unit: str = "") -> None:
        if any(c in "[]{}():,|" for c in unit):
            raise ValueError("Unit contains forbidden characters")
        self._unit = unit

    @property
    def unit(self) -> str:
        """Unit to be appended to the number."""
        return self._unit

    def __eq__(self, other: object) -> bool:
        return isinstance(other, RpcTypeDouble) and self._unit == other._unit

    def __str__(self) -> str:
        return f"f{self._unit}"

    @staticmethod
    def validate(value: SHVType) -> typing.TypeGuard[float]:  # noqa: D102
        return isinstance(value, float)


rpctype_double = RpcTypeDouble()
"""The singleton for :class:`RpcTypeDouble`."""



"""SHV RPC type matching any SHV value."""

from __future__ import annotations

import typing

from .. import SHVType
from .base import RpcType


class RpcTypeAny(RpcType):
    """The Any type representation."""

    __obj = None

    def __new__(cls, alias: str = "") -> RpcTypeAny:  # noqa: D102
        if alias:
            return super().__new__(cls)
        if cls.__obj is None:
            cls.__obj = object.__new__(cls)
        return cls.__obj

    def __init__(self, alias: str = "") -> None:
        self._alias = alias

    @property
    def alias(self) -> str:
        """The alias of this any type if provided."""
        return self._alias

    def __eq__(self, other: object) -> bool:
        return isinstance(other, RpcTypeAny) and self._alias == other._alias

    def __str__(self) -> str:
        return f"?{f'({self._alias})' if self._alias else ''}"

    @staticmethod
    def validate(value: SHVType) -> typing.TypeGuard[SHVType]:  # noqa: D102
        return True


rpctype_any = RpcTypeAny()
"""The singleton for :class:`RpcTypeAny`."""



"""Connection over TCP/IP."""

from __future__ import annotations

import asyncio
import collections.abc
import logging
import ssl

from .abc import RpcClient
from .stream import (
    RpcClientStream,
    RpcProtocolBlock,
    RpcProtocolSerial,
    RpcServerStream,
    RpcTransportProtocol,
)

logger = logging.getLogger(__name__)


class RpcClientTCP(RpcClientStream):
    """RPC connection to some SHV peer over TCP/IP."""

    def __init__(
        self,
        location: str = "localhost",
        port: int = 3755,
        protocol: type[RpcTransportProtocol] = RpcProtocolBlock,
        ssl: ssl.SSLContext | None = None,
    ) -> None:
        super().__init__(protocol)
        self.location = location
        self.port = port
        self.ssl = ssl

    def __str__(self) -> str:
        proto = "tcp" if self.ssl is None else "ssl"
        if self.protocol is RpcProtocolSerial:
            proto += "s"
        location = f"[{self.location}]" if ":" in self.location else self.location
        return f"{proto}:{location}:{self.port}"

    async def _open_connection(
        self,
    ) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        return await asyncio.open_connection(self.location, self.port, ssl=self.ssl)

    @property
    def secure(self) -> bool:  # noqa: D102
        return self.ssl is not None


class RpcServerTCP(RpcServerStream):
    """RPC server listenting for SHV connections in TCP/IP."""

    def __init__(
        self,
        client_connected_cb: collections.abc.Callable[
            [RpcClient], collections.abc.Awaitable[None] | None
        ],
        location: str | None = None,
        port: int = 3755,
        protocol: type[RpcTransportProtocol] = RpcProtocolBlock,
        ssl: ssl.SSLContext | None = None,
    ) -> None:
        super().__init__(client_connected_cb, protocol)
        self.location = location
        self.port = port
        self.ssl = ssl

    def __str__(self) -> str:
        proto = "tcp" if self.ssl is None else "ssl"
        if self.protocol is RpcProtocolSerial:
            proto += "s"
        location = (
            "locahost"
            if self.location is None
            else f"[{self.location}]"
            if ":" in self.location
            else self.location
        )
        return f"server.{proto}:{location}:{self.port}"

    async def _create_server(self) -> asyncio.Server:
        # Note: The hack for '::' here is to allow really bind to all
        # interfaces as that is what this should do and not just loopback.
        return await asyncio.start_server(
            self._client_connect,
            host=None if self.location == "::" else self.location,
            port=self.port,
            ssl=self.ssl,
        )

    async def listen(self) -> None:  # noqa: D102
        was_listening = self.is_serving()
        await super().listen()
        if not was_listening:
            logger.debug("%s: Listening for clients", self)

    def close(self) -> None:  # noqa: D102
        was_listening = self.is_serving()
        super().close()
        if was_listening and (self._server is None or not self._server.is_serving()):
            logger.debug("%s: No longer listening for clients", self)

    async def _client_connect(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        peername = writer.get_extra_info("peername")
        location = f"[{peername[0]}]" if ":" in peername[0] else peername[0]
        logger.debug("%s: New client %s:%d", self, location, peername[1])
        await super()._client_connect(reader, writer)

    class Client(RpcServerStream.Client):
        """RPC client for TCP server connection."""

        def __str__(self) -> str:
            assert isinstance(self._server, RpcServerTCP)
            proto = "tcp" if self._server.ssl is None else "ssl"
            if self.protocol is RpcProtocolSerial:
                proto += "s"
            peername = self._writer.get_extra_info("peername")
            location = f"[{peername[0]}]" if ":" in peername[0] else peername[0]
            return f"tcp:{location}:{peername[1]}"

        @property
        def secure(self) -> bool:  # noqa: D102
            assert isinstance(self._server, RpcServerTCP)
            return self._server.ssl is not None



"""Support for type definition in SHV RPC.

The support consists of the way to transform string representation from and to
python objects. These objects can be used to reason with provided type hint
info. They also support validation of the :class:`shv.SHVType`.
"""

from .any import RpcTypeAny, rpctype_any
from .base import RpcType
from .bitfield import RpcTypeBitfield
from .blob import RpcTypeBlob, rpctype_blob
from .bool import RpcTypeBool, rpctype_bool
from .datetime import RpcTypeDateTime, rpctype_datetime
from .decimal import RpcTypeDecimal, rpctype_decimal
from .double import RpcTypeDouble, rpctype_double
from .enum import RpcTypeEnum
from .imap import RpcTypeIMap, rpctype_imap
from .integer import RpcTypeInteger, rpctype_integer
from .keystruct import RpcTypeKeyStruct
from .list import RpcTypeList, rpctype_list
from .map import RpcTypeMap, rpctype_map
from .null import RpcTypeNull, rpctype_null
from .oneof import RpcTypeOneOf, RpcTypeOptional
from .parse import RpcTypeParseError, rpctype_parse
from .standard import (
    RpcTypeStandard,
    rpctype_alert,
    rpctype_clientinfo,
    rpctype_dir,
    rpctype_exchange_p,
    rpctype_exchange_r,
    rpctype_exchange_v,
    rpctype_getlog_p,
    rpctype_getlog_r,
    rpctype_history_records,
    rpctype_stat,
)
from .string import RpcTypeString, rpctype_string
from .struct import RpcTypeStruct
from .tuple import RpcTypeTuple
from .unsigned import RpcTypeUnsigned, rpctype_unsigned

__all__ = [
    "RpcType",
    "RpcTypeAny",
    "RpcTypeBitfield",
    "RpcTypeBlob",
    "RpcTypeBool",
    "RpcTypeDateTime",
    "RpcTypeDecimal",
    "RpcTypeDouble",
    "RpcTypeEnum",
    "RpcTypeIMap",
    "RpcTypeInteger",
    "RpcTypeKeyStruct",
    "RpcTypeList",
    "RpcTypeMap",
    "RpcTypeNull",
    "RpcTypeOneOf",
    "RpcTypeOptional",
    "RpcTypeParseError",
    "RpcTypeStandard",
    "RpcTypeStandard",
    "RpcTypeString",
    "RpcTypeStruct",
    "RpcTypeTuple",
    "RpcTypeUnsigned",
    "rpctype_alert",
    "rpctype_any",
    "rpctype_blob",
    "rpctype_bool",
    "rpctype_clientinfo",
    "rpctype_datetime",
    "rpctype_decimal",
    "rpctype_dir",
    "rpctype_double",
    "rpctype_exchange_p",
    "rpctype_exchange_r",
    "rpctype_exchange_v",
    "rpctype_getlog_p",
    "rpctype_getlog_r",
    "rpctype_history_records",
    "rpctype_imap",
    "rpctype_integer",
    "rpctype_list",
    "rpctype_map",
    "rpctype_null",
    "rpctype_parse",
    "rpctype_stat",
    "rpctype_string",
    "rpctype_unsigned",
]



"""RPC client that represents device in a physical sense."""

from __future__ import annotations

import collections.abc
import itertools
import time
import typing

from ..rpcdef import RpcAccess, RpcDir, RpcNotImplementedError
from ..rpcdef.alert import RpcAlert
from ..rpcmessage import RpcMessage
from ..value import SHVType
from .base import SHVBase
from .client import SHVClient


class SHVDevice(SHVClient):
    """SHV client that represents a physical device."""

    APP_NAME: str = "pyshv-device"

    DEVICE_NAME: str = "unknown"
    """String identifier for the physical device."""

    DEVICE_VERSION: str = "unknown"
    """The physical device design version."""

    DEVICE_SERIAL_NUMBER: str | None = None
    """Serial number of the physical device."""

    DEVICE_ALERTS: bool = True
    """Control if alerts will be available in the SHV Device."""

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:  # noqa ANNN401
        self._alerts: list[RpcAlert] = []
        super().__init__(*args, **kwargs)  # notype

    @property
    def alerts(self) -> collections.abc.Iterator[RpcAlert]:
        """Iterate over all current alerts."""
        return iter(self._alerts)

    async def change_alerts(
        self,
        add: RpcAlert | collections.abc.Collection[RpcAlert] = tuple(),
        rem: RpcAlert | collections.abc.Collection[RpcAlert] = tuple(),
    ) -> None:
        """Add (or update) given alert.

        :param alert: The alert to be added to the list. Any previous alert with
          same level is replaced.
        """
        if isinstance(add, RpcAlert):
            add = (add,)
        if isinstance(rem, RpcAlert):
            rem = (rem,)
        if not add and not rem:
            return
        self._alerts = list(
            itertools.chain(
                (old for old in self._alerts if old not in add and old not in rem),
                iter(add),
            )
        )
        self._alerts.sort(key=lambda a: a.id)
        await self._send(
            RpcMessage.signal(
                ".device/alerts", value=[alert.value for alert in self._alerts]
            )
        )

    async def _method_call(self, request: SHVBase.Request) -> SHVType:
        match request.path, request.method:
            case ".device", "name":
                return self.DEVICE_NAME
            case ".device", "version":
                return self.DEVICE_VERSION
            case ".device", "serialNumber" if request.access >= RpcAccess.READ:
                return self.DEVICE_SERIAL_NUMBER
            case ".device", "uptime":
                return int(time.monotonic())
            case ".device", "reset" if request.access >= RpcAccess.COMMAND:
                self._device_reset()
                return None  # pragma: no cover
            case ".device/alerts", "get" if (
                self.DEVICE_ALERTS and request.access >= RpcAccess.READ
            ):
                return [alert.value for alert in self._alerts]
        return await super()._method_call(request)

    def _ls(self, path: str) -> collections.abc.Iterator[str]:
        yield from super()._ls(path)
        match path:
            case "":
                yield ".device"
            case ".device" if self.DEVICE_ALERTS:
                yield "alerts"

    def _dir(self, path: str) -> collections.abc.Iterator[RpcDir]:
        yield from super()._dir(path)
        match path:
            case ".device":
                yield RpcDir.getter("name", "n", "s", access=RpcAccess.BROWSE)
                yield RpcDir.getter("version", "n", "s", access=RpcAccess.BROWSE)
                yield RpcDir.getter("serialNumber", "n", "s|n", access=RpcAccess.BROWSE)
                yield RpcDir.getter("uptime", "n", "u|n", access=RpcAccess.BROWSE)
                yield RpcDir("reset", access=RpcAccess.COMMAND)
            case ".device/alerts" if self.DEVICE_ALERTS:
                yield RpcDir.getter(result="[!alert]", signal=True)

    def _device_reset(self) -> None:  # noqa: PLR6301
        """Trigger called to perform the device reset."""
        raise RpcNotImplementedError



"""The SHV RPC access levels."""

from __future__ import annotations

import enum
import functools


class RpcAccess(enum.IntEnum):
    """Method access level."""

    BROWSE = 1
    READ = 8
    WRITE = 16
    COMMAND = 24
    CONFIG = 32
    SERVICE = 40
    SUPER_SERVICE = 48
    DEVEL = 56
    ADMIN = 63

    @classmethod
    @functools.cache
    def strmap(cls) -> dict[str, RpcAccess]:
        """Map from string to this enum."""
        return {
            "bws": cls.BROWSE,
            "rd": cls.READ,
            "wr": cls.WRITE,
            "cmd": cls.COMMAND,
            "cfg": cls.CONFIG,
            "srv": cls.SERVICE,
            "ssrv": cls.SUPER_SERVICE,
            "dev": cls.DEVEL,
            "su": cls.ADMIN,
        }

    @classmethod
    @functools.cache
    def strrmap(cls) -> dict[int, str]:
        """Map from this enum to string."""
        return {v.value: k for k, v in cls.strmap().items()}

    def tostr(self) -> str:
        """Convert to string representation."""
        return self.strrmap().get(self.value, "bws")

    @classmethod
    def fromstr(cls, access: str) -> RpcAccess:
        """Convert to string representation."""
        return cls.strmap().get(access, cls.BROWSE)



"""The loader of the version for source file only installation.

It looks for pyproject.toml file in upper directory and loads version from
there.

This file is overwritten during distribution build to contain only constant
VERSION. Please see __setuptools__.py for details.
"""

import pathlib
import tomllib


def _get_version() -> str:
    pyproject = pathlib.Path(__file__).parents[1] / "pyproject.toml"
    if pyproject.exists():
        with pyproject.open("rb") as f:
            res = tomllib.load(f)["project"]["version"]
            if isinstance(res, str):
                return res
    return "unknown"


VERSION = _get_version()



"""SHV RPC URL used to specify connection and listen settings."""

from __future__ import annotations

import collections.abc
import dataclasses
import enum
import functools
import getpass
import pathlib
import ssl
import typing
import urllib.parse

from .rpclogin import RpcLogin, RpcLoginType


@functools.lru_cache
def _get_user() -> str:
    # getuser fails if there is no account assigned to the UID in the system
    try:
        return getpass.getuser()
    except KeyError:
        return "none"


class RpcProtocol(enum.Enum):
    """Enum of supported RPC protocols by this Python implementation."""

    TCP = enum.auto()
    """TCP/IP with messages transported using Stream transport layer."""
    TCPS = enum.auto()
    """TCP/IP with messages transported using Serial transport layer."""
    SSL = enum.auto()
    """TLS TCP/IP with messages transported using Stream transport layer."""
    SSLS = enum.auto()
    """TLS TCP/IP with messages transported using Serial transport layer."""
    UNIX = enum.auto()
    """Unix local domain named socket using Stream transport layer."""
    UNIXS = enum.auto()
    """Unix local domain named socket using Reliable Serial transport layer."""
    TTY = enum.auto()
    """Serial transport layer."""
    WS = enum.auto()
    """WebSockets transport layer."""
    WSS = enum.auto()
    """WebSockets secure transport layer."""
    CAN = enum.auto()
    """CAN Bus (socketcan) transport layer."""


@dataclasses.dataclass
class RpcUrl:
    """SHV RPC URL specifier.

    This is unified locator for SHV RPC connections that is used to specify
    connections. It is implemented as :func:`dataclasses.dataclass` and you
    can set properties directly or you can parse string URL.

    The URL format is defined in `SHV standard
    <https://silicon-heaven.github.io/shv-doc/rpcurl.html>`_.
    """

    # URL primary fields
    location: str
    """Hostname of the SHV RPC server or path to the socket."""
    port: int = -1
    """Port the SHV RPC server is listening on.

    In case of CAN Bus this is CAN address (0-127).
    """
    protocol: RpcProtocol = RpcProtocol.TCP
    """SHV RPC protocol used to communicate (This is scheme in URL therminology)"""
    login: RpcLogin = dataclasses.field(default_factory=RpcLogin)
    """Parameters for RPC Broker login."""

    # SSL
    ca: str | pathlib.Path | None = None
    """CA certificates."""
    cert: str | pathlib.Path | None = None
    """Certificate."""
    key: str | pathlib.Path | None = None
    """Secret part of client certificate."""
    crl: str | pathlib.Path | None = None
    """Certificate revocation list for server."""
    verify: bool | None = None
    """If peer verifies is required.

    For server this enforces clients verification (clients need their own
    certificate).

    For client this enforces server verification.

    The default is different between server and client. It is ``true`` for
    client and ``false`` for server and thus default is ``None``.
    """

    # TTY
    baudrate: int = 115200
    """Baudrate used for some of the link protocols."""

    # CAN
    can_address: int | None = None
    """The local address (1-63) used by CAN client."""

    default_port: typing.ClassVar[dict[RpcProtocol, int]] = {
        RpcProtocol.TCP: 3755,
        RpcProtocol.TCPS: 3765,
        RpcProtocol.SSL: 3756,
        RpcProtocol.SSLS: 3766,
    }

    def __post_init__(self) -> None:
        """Deduce the correct value for port."""
        if self.port == type(self).port:
            self.port = self.default_port.get(self.protocol, self.port)

    def __str__(self) -> str:
        return self.to_url()

    def ssl_server(self) -> ssl.SSLContext:
        """Create :class:`ssl.SSLContext` for server."""
        if self.protocol not in {RpcProtocol.SSL, RpcProtocol.SSLS}:
            raise ValueError("Not supported for this protocol")
        if self.ca is None:
            raise ValueError("'ca' must be provided")
        if self.cert is None:
            raise ValueError("'cert' must be provided")
        if self.key is None:
            raise ValueError("'key' must be provided")
        context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        context.load_cert_chain(certfile=self.cert, keyfile=self.key)
        context.load_verify_locations(cafile=self.ca)
        if self.crl is not None:
            context.load_verify_locations(cafile=self.crl)
            context.verify_flags = ssl.VERIFY_CRL_CHECK_LEAF
        context.verify_mode = (
            ssl.CERT_REQUIRED if self.verify is True else ssl.CERT_OPTIONAL
        )
        return context

    def ssl_client(self) -> ssl.SSLContext:
        """Create :class:`ssl.SSLContext` for client."""
        if self.protocol not in {RpcProtocol.SSL, RpcProtocol.SSLS}:
            raise ValueError("Not supported for this protocol")
        cacheck = self.verify is not False
        context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        if cacheck:
            if self.ca is None:
                raise ValueError("'ca' must be provided")
            context.load_verify_locations(cafile=self.ca)
        if self.cert is not None and self.key is not None:
            context.load_cert_chain(certfile=self.cert, keyfile=self.key)
        context.check_hostname = cacheck
        context.verify_mode = ssl.CERT_REQUIRED if cacheck else ssl.CERT_NONE
        return context

    @classmethod
    def parse(cls, url: str) -> typing.Self:
        """Parse string URL to the object representation.

        :param url: URL in string format.
        :return: New :class:`RpcUrl` instance.
        :raise ValueError: when invalid URL is passed.
        """
        sr = urllib.parse.urlsplit(url, scheme="unix", allow_fragments=False)
        pqs = urllib.parse.parse_qs(sr.query)

        protocols = {
            "tcp": RpcProtocol.TCP,
            "tcps": RpcProtocol.TCPS,
            "ssl": RpcProtocol.SSL,
            "ssls": RpcProtocol.SSLS,
            "unix": RpcProtocol.UNIX,
            "unixs": RpcProtocol.UNIXS,
            "serial": RpcProtocol.TTY,
            "tty": RpcProtocol.TTY,
            "ws": RpcProtocol.WS,
            "wss": RpcProtocol.WSS,
            "can": RpcProtocol.CAN,
        }
        if sr.scheme not in protocols:
            raise ValueError(f"Invalid scheme: {sr.scheme}")
        protocol = protocols[sr.scheme]

        res = cls("", protocol=protocol)

        res.login.username = sr.username or res.login.username
        match protocol:
            case (
                RpcProtocol.TCP | RpcProtocol.TCPS | RpcProtocol.SSL | RpcProtocol.SSLS
            ):
                res.location = sr.hostname or ""
                if sr.port is not None:
                    res.port = int(sr.port)
                if sr.path:
                    raise ValueError(f"Path not supported for {sr.scheme}: {sr.path}")
            case RpcProtocol.UNIX | RpcProtocol.UNIXS | RpcProtocol.TTY:
                res.location = f"/{sr.netloc}{sr.path}" if sr.netloc else sr.path
            case RpcProtocol.WS | RpcProtocol.WSS:
                if sr.path:
                    if sr.port is not None:
                        raise ValueError(
                            "Port can't be specified with path for websockets"
                        )
                    res.location = f"{sr.netloc}{sr.path}"
                else:
                    res.location = sr.hostname or ""
                    if sr.port is not None:
                        res.port = int(sr.port)
            case RpcProtocol.CAN:
                res.location = sr.hostname or ""
                if sr.port is None:
                    raise ValueError("CAN server address must be specified")
                res.port = int(sr.port)
                if not 0 <= res.port < 128:
                    raise ValueError(f"Invalid CAN server address: {res.port}")
                if sr.path:
                    raise ValueError(
                        f"Path is not supported for {sr.scheme}: {sr.path}"
                    )
            case protocol:
                raise NotImplementedError(f"Protocol: {protocol}")  # pragma: no cover

        res._parse_query(pqs)
        if pqs:
            raise ValueError(f"Unsupported URL queries: {', '.join(pqs.keys())}")

        return res

    def _parse_query(self, pqs: dict[str, list[str]]) -> None:
        """Parse queries.

        This is called from :meth:`parse` to parse query part of the URL. It
        can be extended to get support for additional query parameters.

        The implementation should remove keys that were processed. The valid
        URL thus will end up with empty ``pqs`` after this method returns.

        Also see :meth:`_serialize_query`.

        :param pqs: The query parameters.
        """
        if opts := pqs.pop("token", []):
            self.login.token = opts[0]
            self.login.login_type = RpcLoginType.TOKEN
        else:
            if opts := pqs.pop("user", []):
                self.login.username = opts[0]
            if opts := pqs.pop("shapass", []):
                if len(opts[0]) != 40:
                    raise ValueError("SHA1 password must have 40 characters.")
                self.login.password = opts[0]
                self.login.login_type = RpcLoginType.SHA1
                # We prefer SHA1 password and thus discard plain if both are present
                pqs.pop("password", [])
            elif opts := pqs.pop("password", []):
                self.login.password = opts[0]
                self.login.login_type = RpcLoginType.PLAIN
        if opts := pqs.pop("devid", []):
            self.login.device_id = opts[0]
        if opts := pqs.pop("devmount", []):
            self.login.device_mount_point = opts[0]
        if self.protocol in {RpcProtocol.SSL, RpcProtocol.SSLS}:
            if opts := pqs.pop("ca", []):
                self.ca = opts[0]
            if opts := pqs.pop("cafile", []):
                self.ca = pathlib.Path(opts[0])
            if opts := pqs.pop("cert", []):
                self.cert = opts[0]
            if opts := pqs.pop("certfile", []):
                self.cert = pathlib.Path(opts[0])
            if opts := pqs.pop("key", []):
                self.key = opts[0]
            if opts := pqs.pop("keyfile", []):
                self.key = pathlib.Path(opts[0])
            if opts := pqs.pop("crl", []):
                self.crl = opts[0]
            if opts := pqs.pop("crlfile", []):
                self.crl = pathlib.Path(opts[0])
            if opts := pqs.pop("verify", []):
                if opts[0] in {"true", "t", "y"}:
                    self.verify = True
                elif opts[0] in {"false", "f", "n"}:
                    self.verify = False
                else:
                    raise ValueError(f"Invalid for verify: {opts[0]}")
        if self.protocol is RpcProtocol.TTY:
            if opts := pqs.pop("baudrate", []):
                self.baudrate = int(opts[0])
        if self.protocol is RpcProtocol.CAN:
            if opts := pqs.pop("caddr", []):
                self.can_address = int(opts[0])
                if not 0 <= self.can_address < 128:
                    raise ValueError("CAN address must be in range <0,127>")

    def to_url(self, public: bool = False) -> str:
        """Convert to string URL.

        :param public: You can pass ``True`` to not include login credentials.
        :returns: The string representation of the RPC URL.
        """
        protocols = {
            RpcProtocol.TCP: "tcp",
            RpcProtocol.TCPS: "tcps",
            RpcProtocol.SSL: "ssl",
            RpcProtocol.SSLS: "ssls",
            RpcProtocol.UNIX: "unix",
            RpcProtocol.UNIXS: "unixs",
            RpcProtocol.TTY: "serial",
            RpcProtocol.WS: "ws",
            RpcProtocol.WSS: "wss",
            RpcProtocol.CAN: "can",
        }
        user_added = not self.login.username or self.login.username == RpcLogin.username
        match self.protocol:
            case (
                RpcProtocol.TCP | RpcProtocol.TCPS | RpcProtocol.SSL | RpcProtocol.SSLS
            ):
                netloc = "//"
                if not user_added:
                    netloc += f"{self.login.username}@"
                    user_added = True
                if ":" in self.location:
                    netloc += f"[{self.location}]"
                else:
                    netloc += self.location
                if self.port != self.default_port[self.protocol]:
                    netloc += f":{self.port}"
            case RpcProtocol.UNIX | RpcProtocol.UNIXS | RpcProtocol.TTY:
                netloc = self.location
            case RpcProtocol.WS | RpcProtocol.WSS:
                if self.port == type(self).port:
                    netloc = self.location
                else:
                    netloc = "//"
                    if not user_added:
                        netloc += f"{self.login.username}@"
                        user_added = True
                    if ":" in self.location:
                        netloc += f"[{self.location}]"
                    else:
                        netloc += self.location
                    netloc += f":{self.port}"
            case RpcProtocol.CAN:
                netloc = f"//{self.location}:{self.port}"
            case protocol:
                raise NotImplementedError(f"Protocol: {protocol}")  # pragma: no cover

        opts = [] if user_added else [f"user={urllib.parse.quote(self.login.username)}"]
        opts += self._serialize_query(public)
        # TODO correctly escape the URL
        return (
            f"{protocols[self.protocol]}:{netloc}{'?' if opts else ''}{'&'.join(opts)}"
        )

    def _serialize_query(self, public: bool) -> collections.abc.Iterator[str]:
        """Serialize queries.

        This is called from :meth:`to_url` to serialize query part of the URL.
        It can be extended to get support for additional query parameters.

        Also see :meth:`_parse_query`.

        :param pqs: The query parameters.
        """
        if self.login.device_id:
            yield f"devid={urllib.parse.quote(self.login.device_id)}"
        if self.login.device_mount_point:
            yield f"devmount={urllib.parse.quote(self.login.device_mount_point)}"
        if not public:
            if self.login.token:
                yield f"token={self.login.token}"
            if self.login.password:
                if self.login.login_type is RpcLoginType.SHA1:
                    yield f"shapass={self.login.password}"
                elif self.login.login_type is RpcLoginType.PLAIN:
                    yield f"password={urllib.parse.quote(self.login.password)}"
                else:
                    raise NotImplementedError  # pragma: no cover
        if self.baudrate != type(self).baudrate:
            yield f"baudrate={self.baudrate}"
        if self.can_address != type(self).can_address:
            yield f"caddr={self.can_address}"
        if self.ca != type(self).ca:
            yield f"{'cafile' if isinstance(self.ca, pathlib.Path) else 'ca'}={self.ca}"
        if self.cert != type(self).cert:
            yield f"{'certfile' if isinstance(self.cert, pathlib.Path) else 'cert'}={self.cert}"
        if self.key != type(self).key:
            yield f"{'keyfile' if isinstance(self.key, pathlib.Path) else 'key'}={self.key}"
        if self.crl != type(self).crl:
            yield f"{'crlfile' if isinstance(self.crl, pathlib.Path) else 'crl'}={self.crl}"
        if self.verify != type(self).verify:
            yield f"verify={'true' if self.verify else 'false'}"



"""Connection over CAN bus."""

from __future__ import annotations

import asyncio
import collections.abc
import concurrent
import dataclasses
import enum
import errno
import logging
import random
import time
import typing
import weakref

from .__canlog import patchlogger as __patchlogger
from .abc import RpcClient, RpcServer

logger = logging.getLogger(__name__)

try:
    import can
    import can.interfaces.socketcan

    CAN_IMPORT = None
    __patchlogger(logging.getLogger("can"))
    __patchlogger(logging.getLogger("can.interfaces.socketcan.socketcan"))
    __patchlogger(logging.getLogger("can.interfaces.socketcan.socketcan.tx"))
    __patchlogger(logging.getLogger("can.interfaces.socketcan.socketcan.rx"))
except ImportError as exc:
    CAN_IMPORT = exc


class SHVCAN:
    """The interface used to manage SHV CAN-FD communication."""

    CANID_SHVCAN: typing.Final = 0x400
    CANID_UNUSED: typing.Final = 0x200
    CANID_FIRST_MASK: typing.Final = 0x100
    CANID_ADDRESS_MASK: typing.Final = 0x0FF

    class DLC(enum.IntEnum):
        """DLC used in RTR frames."""

        ADDR_ACQ = 0
        """Address acquistion."""
        ADDR_ANC_LISTEN = 1
        """Address announce for peers accepting connections."""
        ADDR_ANC_IGN = 2
        """Address announce for peers not accepting new connections."""
        ADDR_DISC_LISTEN = 5
        """Address discovery for peers accepting connections."""
        ADDR_DISC_IGN = 6
        """Address discovery for peers not accepting new connections."""
        ADDR_DISC = 7
        """Address discovery for all peers."""

    class AddressAcquisitionError(RuntimeError):
        """Failure to acquire dynamic address."""

        def __init__(self, acq_conflict: bool) -> None:
            super().__init__(
                "Dynamic address acquisition collide with other one"
                if acq_conflict
                else "Dynamic address already taken, can't acquire",
                acq_conflict,
            )

        @property
        def acq_conflict(self) -> bool:
            """Check if this is acquisition or existing assignment conflict."""
            return bool(self.args[1])

    @dataclasses.dataclass
    class Local:
        """Local address."""

        shvcan: SHVCAN
        address: int

        _peers: weakref.WeakValueDictionary[int, SHVCAN.Peer] = dataclasses.field(
            default_factory=weakref.WeakValueDictionary
        )

        def __post_init__(self) -> None:
            if self.address in self.shvcan._locals:
                raise ValueError(f"Local SHVCAN address {self.address} already used")
            self.shvcan._locals[self.address] = self
            self._new_peer_callback: (
                collections.abc.Callable[
                    [SHVCAN.Peer], collections.abc.Awaitable[None] | None
                ]
                | None
            ) = None
            self._dynfuture: asyncio.Future[bool] | None = (
                asyncio.get_running_loop().create_future()
                if 0x80 <= self.address <= 0xFF
                else None
            )

        @property
        def active(self) -> bool:
            """Check if this peer is still tied to :class:`SHVCAN`."""
            return self.shvcan._locals.get(self.address) is self

        async def announce(self) -> None:
            """Send announce."""
            await self.shvcan._send_frames((
                self._rtrframe(
                    SHVCAN.DLC.ADDR_ANC_LISTEN
                    if self.listening
                    else SHVCAN.DLC.ADDR_ANC_IGN
                ),
            ))

        async def discover(
            self, listening: bool = True, notlistening: bool = True
        ) -> None:
            """Request discovery of the active addresses on the CAN Bus."""
            dlc = {
                (True, False): SHVCAN.DLC.ADDR_DISC_LISTEN,
                (False, True): SHVCAN.DLC.ADDR_DISC_IGN,
                (True, True): SHVCAN.DLC.ADDR_DISC,
            }.get((listening, notlistening), None)
            if dlc is not None:
                await self.shvcan._send_frames((self._rtrframe(dlc),))

        @property
        def address_acquired(self) -> bool:
            """Check if address this local has is acquired by us."""
            return self._dynfuture is None

        async def dynamic_address_acquire(self) -> None:
            """Perform dynamic address acquistion.

            This has to be called at the start for addresses in dynamic address
            range.

            :raise SHVCAN.AddressAcquisitionError: In case address can't be
              acquired.
            """
            if self.active and self._dynfuture is not None:
                for _ in range(8):
                    await self.shvcan._send_frames((
                        self._rtrframe(SHVCAN.DLC.ADDR_ACQ),
                    ))
                    if (await asyncio.wait((self._dynfuture,), timeout=0.05))[0]:
                        assert self._dynfuture.done()
                        del self.shvcan._locals[self.address]
                        raise SHVCAN.AddressAcquisitionError(self._dynfuture.result())
                self._dynfuture = None

        @property
        def listening(self) -> bool:
            """If new connections are being accepted."""
            return self.active and self._new_peer_callback is not None

        async def bind(
            self,
            callback: collections.abc.Callable[
                [SHVCAN.Peer], collections.abc.Awaitable[None] | None
            ],
        ) -> None:
            """Accept new connections."""
            if self._new_peer_callback is not None:
                raise ValueError("Address already bound.")
            if self._dynfuture is not None:
                raise ValueError("Address not yet negotiated")
            self._new_peer_callback = callback
            await self.announce()

        def unbind(self) -> None:
            """Stop listening for new connections."""
            self._new_peer_callback = None

        async def deactivate(self) -> None:
            """Close the local address.

            This won't disconnect any existing connections. You must do that
            on your own.
            """
            if self.active:
                if self._peers:
                    raise RuntimeError("There are still existing connections.")
                del self.shvcan._locals[self.address]
                if self._new_peer_callback is None and not self.shvcan._locals:
                    await self.shvcan.terminate()

        def _rtrframe(self, dlc: int) -> can.Message:
            priority = dlc not in {SHVCAN.DLC.ADDR_ACQ}
            return self._frame(priority, is_remote_frame=True, dlc=dlc)

        def _frame(self, priority: bool, **kwargs: typing.Any) -> can.Message:  # noqa: ANN401
            return can.Message(
                arbitration_id=SHVCAN.CANID_SHVCAN
                + SHVCAN.CANID_UNUSED
                + (0 if priority else SHVCAN.CANID_FIRST_MASK)
                + self.address,
                is_extended_id=False,
                **kwargs,
            )

    @dataclasses.dataclass
    class Peer:
        """Peer representing handle in SHVCAN."""

        local: SHVCAN.Local
        address: int
        queue: asyncio.Queue[bytes | None] = dataclasses.field(
            default_factory=asyncio.Queue
        )

        _sendlock: asyncio.Lock = dataclasses.field(default_factory=asyncio.Lock)
        _scounter: int = dataclasses.field(
            default_factory=lambda: random.randrange(0x7F)  # noqa: S311
        )

        _rdata: bytearray = dataclasses.field(default_factory=bytearray)
        _rcounter: int = 0

        def __post_init__(self) -> None:
            if self.address in self.local._peers:
                raise ValueError(
                    f"Connection between {self.local.address} and {self.address} already used"
                )
            self._acknowledgement: asyncio.Future[None] | None = None
            self.local._peers[self.address] = self

        @property
        def connected(self) -> bool:
            """Check if this peer is still tied to :class:`SHVCAN`."""
            return self.local._peers.get(self.address) is self

        async def send(self, msg: bytes) -> None:
            """Send given message to this peer."""
            msglen = len(msg)
            if msglen == 0:
                raise ValueError("Message can't have zero length")
            first_block = msg[0:62]
            async with self._sendlock:
                frame = self._dataframe(first_block, first=True, last=msglen <= 62)
                future = asyncio.get_running_loop().create_future()
                self._acknowledgement = future
                try:
                    for _ in range(25):
                        await self.local.shvcan._send_frames((frame,))
                        if (await asyncio.wait((future,), timeout=0.2))[0]:
                            break
                    else:
                        raise EOFError("No acknowledgement in time")
                finally:
                    self._acknowledgement = None

            await self.local.shvcan._send_frames(
                self._dataframe(msg[i : i + 62], last=(i + 62) > msglen)
                for i in range(62, msglen, 62)
            )

        async def disconnect(self) -> None:
            """Close the connection to the peer."""
            if self.connected:
                del self.local._peers[self.address]
                await self.local.shvcan._send_frames((self._disconnectframe(),))
                if self.local._new_peer_callback is None and not self.local._peers:
                    await self.local.deactivate()

        def _dataframe(
            self, data: bytes, first: bool = False, last: bool = False
        ) -> can.Message:
            assert 0 < len(data) <= 62
            self._scounter = (self._scounter + 1) % 0x80
            self._scounter += 0x80 if last else 0
            return self.local._frame(
                not first, is_fd=True, data=bytes((self.address, self._scounter)) + data
            )

        def _acknowledgementframe(self) -> can.Message:
            return self.local._frame(
                True, is_fd=True, data=bytes((self.address, self._rcounter))
            )

        def _disconnectframe(self) -> can.Message:
            return self.local._frame(False, is_fd=True, data=bytes((self.address,)))

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:  # noqa: ANN401
        if CAN_IMPORT is not None:
            raise CAN_IMPORT
        self._args = args
        self._kwargs = kwargs
        self._locals: weakref.WeakValueDictionary[int, SHVCAN.Local]
        self._locals = weakref.WeakValueDictionary()
        self.bus = can.Bus(*self._args, **self._kwargs)
        if self.bus.protocol is not can.CanProtocol.CAN_FD:
            self.bus.shutdown()
            raise ValueError("Only CAN-FD is supported")
        self._notifier = can.Notifier(
            self.bus,
            (self._receive_frame,),
            timeout=0.05,
            loop=asyncio.get_running_loop(),
        )
        self._sendexecutor = concurrent.futures.ThreadPoolExecutor(1)
        self._discovery_cbs: set[collections.abc.Callable[[int, bool], None]] = set()

    def __str__(self) -> str:
        if isinstance(self.bus, can.interfaces.socketcan.SocketcanBus):
            return self.bus.channel
        else:
            return self.bus.channel_info

    def existing_local(self, address: int) -> SHVCAN.Local:
        """Get existing :class:`SHVCAN.Local` for given address."""
        return self._locals[address]

    def static_local(self, address: int) -> SHVCAN.Local:
        """Create new static local address.

        :param address: Static local address.
        :raise ValueError: In case address is not in correct range for static
          address or if address is already in use.
        """
        if not 0x00 <= address <= 0x7F:
            raise ValueError("Address must be in range from 0x0 to 0x7F")
        return self.Local(self, address)

    async def dynamic_local(self) -> SHVCAN.Local:
        """Negotiate a dynamic local address to be used.

        :return: New dynamic local address object.
        :raise RuntimeError: If there is no free dynamic address available.
        """
        # TODO possibly also exclude those we are already communicating with
        available = [v for v in range(0x80, 0x100) if v not in self._locals]
        while available:
            local = self.Local(self, random.choice(available))  # noqa: S311
            try:
                await local.dynamic_address_acquire()
            except self.AddressAcquisitionError as exc:
                if not exc.acq_conflict:
                    available.remove(local.address)
            else:
                return local
        raise RuntimeError("No free dynamic address is available")

    def register_discovery_callback(
        self, cb: collections.abc.Callable[[int, bool], None]
    ) -> None:
        """Register given callback to be called when address is announced."""
        self._discovery_cbs.add(cb)

    def unregister_discovery_callback(
        self, cb: collections.abc.Callable[[int, bool], None]
    ) -> None:
        """Remove previous added discovery callback with :meth:`register_discovery_callback`."""
        self._discovery_cbs.remove(cb)

    async def terminate(self) -> None:
        """Terminate the CAN instance."""
        if self._locals:
            raise RuntimeError("There are still existing locals.")
        self._notifier.stop()
        self.bus.shutdown()

    async def _receive_frame(self, msg: can.Message) -> None:
        if (
            msg.is_error_frame  # TODO should we not ignore error frames?
            or msg.is_extended_id
            or not bool(msg.arbitration_id & self.CANID_SHVCAN)
        ):  # Ignore extended ID frames and those with ID not for SHV CANFD
            return
        src = msg.arbitration_id & self.CANID_ADDRESS_MASK
        first = bool(msg.arbitration_id & self.CANID_FIRST_MASK)
        if msg.is_remote_frame:
            match msg.dlc:
                case SHVCAN.DLC.ADDR_ACQ:
                    if (local := self._locals.get(src, None)) is not None:
                        if local._dynfuture is not None and not local._dynfuture.done():
                            local._dynfuture.set_result(True)
                        else:
                            await local.announce()
                case SHVCAN.DLC.ADDR_ANC_LISTEN | SHVCAN.DLC.ADDR_ANC_IGN:
                    if (local := self._locals.get(src, None)) is not None:
                        if local._dynfuture is not None and not local._dynfuture.done():
                            local._dynfuture.set_result(False)
                    for cb in self._discovery_cbs:
                        cb(src, msg.dlc == SHVCAN.DLC.ADDR_ANC_LISTEN)
                case SHVCAN.DLC.ADDR_DISC_LISTEN:
                    for v in self._locals.values():
                        if v.listening:
                            await v.announce()
                case SHVCAN.DLC.ADDR_DISC_IGN:
                    for v in self._locals.values():
                        if not v.listening:
                            await v.announce()
                case SHVCAN.DLC.ADDR_DISC:
                    for v in self._locals.values():
                        await v.announce()
            return

        datalen = len(msg.data)
        if datalen < 1:
            return  # Invalid frame (at least destination must be present)
        local = self._locals.get(msg.data[0])
        if local is None:
            return  # Not for us
        new_peer = False
        peer = local._peers.get(src)
        if peer is None:
            if local._new_peer_callback is None or not first or datalen < 2:
                return  # Not new connection or we don't listen
            new_peer = True
            peer = self.Peer(local, src)
            res = local._new_peer_callback(peer)
            if isinstance(res, collections.abc.Awaitable):
                await res
        match datalen:
            case 1:  # Disconnect
                await peer.queue.put(None)
                del local._peers[src]
                return
            case 2:  # Acknowledgement
                if peer._scounter == msg.data[1] and (ack := peer._acknowledgement):
                    ack.set_result(None)
                return

        last = bool(msg.data[1] & 0x80)
        counter = msg.data[1] & 0x7F
        if first:
            peer._rdata.clear()  # Make sure that we start new message
        elif not peer._rdata or counter == peer._rcounter:
            return  # We don't have start or copy of previous message
        elif counter != ((peer._rcounter + 1) % 0x80):
            return peer._rdata.clear()  # Missed frame or message abort
        peer._rcounter = msg.data[1]
        peer._rdata.extend(msg.data[2:])
        if first:
            await self._send_frames((peer._acknowledgementframe(),))
        if last:
            # We ignore here an initial reset message if this is a new peer.
            if not new_peer or peer._rdata != b"\0":
                await peer.queue.put(
                    bytes(peer._rdata)
                    if len(peer._rdata) <= 8
                    else peer._rdata.rstrip(b"\0")
                )
            peer._rdata.clear()

    async def _send_frames(self, frames: collections.abc.Iterable[can.Message]) -> None:
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(self._sendexecutor, self.__send_frames, frames)

    def __send_frames(self, frames: collections.abc.Iterable[can.Message]) -> None:
        for msg in frames:
            while True:
                try:
                    self.bus.send(msg)
                except can.exceptions.CanOperationError as exc:
                    if f"[Error Code {errno.ENOBUFS}]" in exc.args[0]:
                        time.sleep(0.01)
                    elif "Transmit buffer full" != exc.args[0]:
                        raise exc
                else:
                    break

    @classmethod
    def socketcan(cls, interface: str) -> SHVCAN:
        """Get :class:`SHVCAN` for given socket CAN interface."""
        return cls(channel=interface, interface="socketcan", fd=True)

    @classmethod
    def virtualcan(cls, name: str) -> SHVCAN:
        """Get :class:`SHVCAN` for given virtual interface name.

        This virtual interface should be used primarilly for the testing of the
        CAN functionality. It works only in single process.
        """
        return cls(name, interface="virtual", protocol=can.CanProtocol.CAN_FD)


class RpcClientCAN(RpcClient):
    """RPC client communicating over CAN bus.

    :param location: The interface to be used for SHV CAN or directly
      :class:`SHVCAN` instance.
    :param address: Address of the peer client should connect to.
    :param client_address: Local address to be used. Note that only values
      from ``0`` to ``127`` are allowed. The higher addresses can be used only
      if you pass ``None`` and in such case a random dynamic address is used
      instead.
    """

    def __init__(
        self, location: str | SHVCAN, address: int, client_address: int | None = None
    ) -> None:
        if not 0 <= address < 256:
            raise ValueError(f"Invalid address: {address}")
        if client_address is not None and not 0 <= client_address < 128:
            raise ValueError(f"Invalid static client address: {client_address}")
        super().__init__()
        self._shvcan: SHVCAN = (
            location if isinstance(location, SHVCAN) else SHVCAN.socketcan(location)
        )
        self._address = address
        self._client_address = client_address
        self._peer: SHVCAN.Peer | None = None
        self._disconect_task: asyncio.Task | None = None
        self._disconnect_event = asyncio.Event()

    @property
    def shvcan(self) -> SHVCAN:
        """Access to the associated :class:`SHVCAN`."""
        return self._shvcan

    def __str__(self) -> str:
        return f"can://{self._shvcan}:{self._address}[{self._client_address}]"

    @property
    def address(self) -> int:
        """The CAN address client is connected to."""
        return self._address

    @property
    def client_address(self) -> int | None:
        """The local CAN address assigned to this client."""
        if self._client_address is not None or self._peer is None:
            return self._client_address
        return self._peer.local.address

    async def _send(self, msg: bytes) -> None:
        if self._peer is None:
            raise EOFError("Not connected")
        await self._peer.send(msg)

    async def _receive(self) -> bytes:
        if self._peer is None:
            raise EOFError("Not connected")
        res = await self._peer.queue.get()
        self._peer.queue.task_done()
        if res is None:
            raise EOFError("Peer disconnected")
        return res

    async def reset(self) -> None:  # noqa: D102
        if not self.connected:
            if self._client_address is None:
                local = await self._shvcan.dynamic_local()
            else:
                try:
                    local = self._shvcan.existing_local(self._client_address)
                except KeyError:
                    local = SHVCAN.Local(self._shvcan, self._client_address)
            self._peer = SHVCAN.Peer(local, self._address)
            logger.debug("%s: Connected", self)
        # Always send reset message to establish connection right away
        await super().reset()

    @property
    def connected(self) -> bool:  # noqa: D102
        return self._peer is not None and self._peer.connected

    def _disconnect(self) -> None:
        if self._peer is not None and self._disconect_task is None:
            self._disconnect_task = asyncio.create_task(self._peer.disconnect())
            self._disconnect_event.set()

    async def wait_disconnect(self) -> None:  # noqa D027
        if self._peer is not None:
            await self._disconnect_event.wait()
            await self._disconnect_task


class RpcServerCAN(RpcServer):
    """RPC server listening on CAN bus.

    :param client_connected_cb: Callable that is called when new client is
      connected.
    :param location: The interface to be used for SHV CAN or directly
      :class:`SHVCAN` instance.
    :param address: Static listening address or ``None`` if a random dynamic
      address should be acquired.
    """

    def __init__(
        self,
        client_connected_cb: collections.abc.Callable[
            [RpcClient], collections.abc.Awaitable[None] | None
        ],
        location: str | SHVCAN,
        address: int | None = None,
    ) -> None:
        if address is not None and not 0 <= address < 128:
            raise ValueError(f"Invalid static address: {address}")
        self.client_connected_cb = client_connected_cb
        """Callback that is called when new client is connected."""
        self._shvcan: SHVCAN = (
            location if isinstance(location, SHVCAN) else SHVCAN.socketcan(location)
        )
        """Associated instance of :class:`SHVCAN`."""
        self._address = address
        """The address server is listening on."""
        self._local: SHVCAN.Local | None = None
        self._clients: weakref.WeakSet[RpcServerCAN.Client] = weakref.WeakSet()
        self._close_event = asyncio.Event()

    @property
    def shvcan(self) -> SHVCAN:
        """Access to the associated :class:`SHVCAN`."""
        return self._shvcan

    def __str__(self) -> str:
        return f"server.can://{self._shvcan}:{self._address}"

    @property
    def address(self) -> int | None:
        """The CAN address assigned to this server."""
        if self._address is not None or self._local is None:
            return self._address
        return self._local.address

    def is_serving(self) -> bool:  # noqa: D102
        return self._local is not None and self._local.listening

    async def listen(self) -> None:  # noqa: D102
        if not self.is_serving():
            if self._address is None:
                self._local = await self._shvcan.dynamic_local()
            else:
                self._local = self._shvcan.static_local(self._address)
            await self._local.bind(self._peer_connect)
            self._close_event.clear()
            logger.debug("%s: Listening", self)

    async def listen_forever(self) -> None:  # noqa: D102
        await self.listen()
        await self.wait_terminated()

    async def _peer_connect(self, peer: SHVCAN.Peer) -> None:
        client = self.Client(peer, self)
        self._clients.add(client)
        res = self.client_connected_cb(client)
        if isinstance(res, collections.abc.Awaitable):
            await res

    def close(self) -> None:  # noqa: D102
        if self.is_serving():
            assert self._local is not None
            self._local.unbind()
            self._close_event.set()
            logger.debug("%s: No longer listening", self)

    async def wait_closed(self) -> None:  # noqa: D102
        if self._local is not None:
            await self._close_event.wait()
            if not self._local._peers:
                await self._local.deactivate()
            self._local = None

    def terminate(self) -> None:  # noqa: D102
        self.close()
        for client in self._clients:
            client.disconnect()

    async def wait_terminated(self) -> None:  # noqa: D102
        await self.wait_closed()
        res = await asyncio.gather(
            *(c.wait_disconnect() for c in self._clients),
            return_exceptions=True,
        )
        excs = [v for v in res if isinstance(v, BaseException)]
        if excs:
            if len(excs) == 1:
                raise excs[0]
            raise BaseExceptionGroup("", excs)

    class Client(RpcClient):
        """RPC connection to SHV server peer over CAN bus."""

        def __init__(self, peer: SHVCAN.Peer, server: RpcServerCAN) -> None:
            super().__init__()
            self._peer: SHVCAN.Peer = peer
            self._server = server
            self._disconect_task: asyncio.Task | None = None
            self._disconnect_event = asyncio.Event()

        @property
        def server(self) -> RpcServerCAN:
            """Access to the associated server."""
            return self._server

        def __str__(self) -> str:
            return f"can://{self._server._shvcan}:{self._peer.address}[{self._peer.local.address}]"

        @property
        def address(self) -> int:
            """The CAN address client is connected to."""
            res = self._server.address
            assert res is not None
            return res

        @property
        def client_address(self) -> int:
            """The local CAN address assigned to this client."""
            return self._peer.address

        async def _send(self, msg: bytes) -> None:
            await self._peer.send(msg)

        async def _receive(self) -> bytes:
            res = await self._peer.queue.get()
            self._peer.queue.task_done()
            if res is None:
                raise EOFError("Peer disconnected")
            return res

        @property
        def connected(self) -> bool:  # noqa: D102
            return self._peer.connected

        def _disconnect(self) -> None:
            if self._peer is not None and self._disconect_task is None:
                self._disconnect_task = asyncio.create_task(self._peer.disconnect())
                self._disconnect_event.set()

        async def wait_disconnect(self) -> None:  # noqa D027
            if self._peer is not None:
                await self._disconnect_event.wait()
                await self._disconnect_task
